# This file is part of the MapProxy project.
# Copyright (C) 2011-2013 Omniscale <http://omniscale.de>

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import timezone, datetime
import hashlib
import logging
import os
import re
import threading
import time
from io import BytesIO
from itertools import groupby

from mapproxy.cache.base import TileCacheBase, tile_buffer, REMOVE_ON_UNLOCK
from mapproxy.image import ImageSource
from mapproxy.srs import get_epsg_num
from mapproxy.util.fs import ensure_directory
from mapproxy.util.lock import FileLock
from mapproxy.util.sqlite3 import sqlite3


log = logging.getLogger(__name__)


class GeopackageCache(TileCacheBase):
    supports_timestamp = False

    def __init__(
            self, geopackage_file, tile_grid, table_name, with_timestamps=False, timeout=30, wal=False, coverage=None,
            directory_permissions=None, file_permissions=None):
        super(GeopackageCache, self).__init__(coverage)
        self.tile_grid = tile_grid
        self.table_name = self._check_table_name(table_name)
        md5 = hashlib.new('md5', geopackage_file.encode('utf-8'), usedforsecurity=False)
        self.lock_cache_id = 'gpkg' + md5.hexdigest()
        self.geopackage_file = geopackage_file
        self.directory_permissions = directory_permissions
        self.file_permissions = file_permissions

        if coverage:
            self.bbox = coverage.transform_to(self.tile_grid.srs).bbox
        else:
            self.bbox = self.tile_grid.bbox
        # XXX timestamps not implemented
        self.supports_timestamp = with_timestamps
        self.timeout = timeout
        self.wal = wal
        self.ensure_gpkg()
        self._db_conn_cache = threading.local()

    @property
    def db(self):
        if not getattr(self._db_conn_cache, 'db', None):
            self.ensure_gpkg()
            self._db_conn_cache.db = sqlite3.connect(self.geopackage_file, timeout=self.timeout)
        return self._db_conn_cache.db

    def uncached_db(self):
        return sqlite3.connect(self.geopackage_file, timeout=self.timeout)

    def cleanup(self):
        """
        Close all open connection and remove them from cache.
        """
        if getattr(self._db_conn_cache, 'db', None):
            self._db_conn_cache.db.close()
        self._db_conn_cache.db = None

    @staticmethod
    def _check_table_name(table_name):
        """
        >>> GeopackageCache._check_table_name("test")
        'test'
        >>> GeopackageCache._check_table_name("test_2")
        'test_2'
        >>> GeopackageCache._check_table_name("test-2")
        'test-2'
        >>> GeopackageCache._check_table_name("test3;")
        Traceback (most recent call last):
        ...
        ValueError: The table_name test3; contains unsupported characters.
        >>> GeopackageCache._check_table_name("table name")
        Traceback (most recent call last):
        ...
        ValueError: The table_name table name contains unsupported characters.

        @param table_name: A desired name for an geopackage table.
        @return: The name of the table if it is good, otherwise an exception.
        """
        # Regex string indicating table names which will be accepted.
        regex_str = '^[a-zA-Z0-9_-]+$'
        if re.match(regex_str, table_name):
            return table_name
        else:
            msg = ("The table name may only contain alphanumeric characters, an underscore, "
                   "or a dash: {}".format(regex_str))
            log.info(msg)
            raise ValueError("The table_name {0} contains unsupported characters.".format(table_name))

    def ensure_gpkg(self):
        if not os.path.isfile(self.geopackage_file):
            with FileLock(self.geopackage_file + '.init.lck', remove_on_unlock=REMOVE_ON_UNLOCK,
                          directory_permissions=self.directory_permissions, file_permissions=self.file_permissions):
                ensure_directory(self.geopackage_file, self.directory_permissions)
                self._initialize_gpkg()
        else:
            if not self.check_gpkg():
                ensure_directory(self.geopackage_file, self.directory_permissions)
                self._initialize_gpkg()

    def check_gpkg(self):
        if not self._verify_table():
            return False
        if not self._verify_gpkg_contents():
            return False
        if not self._verify_tile_size():
            return False
        return True

    def _verify_table(self):
        with self.uncached_db() as db:
            cur = db.execute("""SELECT name FROM sqlite_master WHERE type='table' AND name=?""",
                             (self.table_name,))
            content = cur.fetchone()
            if not content:
                # Table doesn't exist _initialize_gpkg will create a new one.
                return False
            return True

    def _verify_gpkg_contents(self):
        with self.uncached_db() as db:
            cur = db.execute("""SELECT * FROM gpkg_contents WHERE table_name = ?""", (self.table_name,))

            results = cur.fetchone()
            if not results:
                # Table doesn't exist in gpkg_contents _initialize_gpkg will add it.
                return False
            gpkg_data_type = results[1]
            gpkg_srs_id = results[9]
            cur = db.execute("""SELECT * FROM gpkg_spatial_ref_sys WHERE srs_id = ?""", (gpkg_srs_id,))

            gpkg_coordsys_id = cur.fetchone()[3]
            if gpkg_data_type.lower() != "tiles":
                log.info("The geopackage table name already exists for a data type other than tiles.")
                raise ValueError("table_name is improperly configured.")
            if gpkg_coordsys_id != get_epsg_num(self.tile_grid.srs.srs_code):
                log.info(
                    f"The geopackage {self.geopackage_file} table name {self.table_name} already exists and has an SRS"
                    f" of {gpkg_coordsys_id}, which does not match the configured Mapproxy SRS of"
                    f" {get_epsg_num(self.tile_grid.srs.srs_code)}.")
                raise ValueError("srs is improperly configured.")
            return True

    def _verify_tile_size(self):
        with self.uncached_db() as db:
            cur = db.execute(
                """SELECT * FROM gpkg_tile_matrix WHERE table_name = ?""",
                (self.table_name,))

            results = cur.fetchall()
            results = results[0]
            tile_size = self.tile_grid.tile_size

            if not results:
                # There is no tile conflict. Return to allow the creation of new tiles.
                return True

            gpkg_table_name, gpkg_zoom_level, gpkg_matrix_width, gpkg_matrix_height, gpkg_tile_width, \
                gpkg_tile_height, gpkg_pixel_x_size, gpkg_pixel_y_size = results
            resolution = self.tile_grid.resolution(gpkg_zoom_level)
            if gpkg_tile_width != tile_size[0] or gpkg_tile_height != tile_size[1]:
                log.info(
                    f"The geopackage {self.geopackage_file} table name {self.table_name} already exists and has tile"
                    f" sizes of ({gpkg_tile_width},{gpkg_tile_height}) which is different than the configure tile sizes"
                    f" of ({tile_size[0]},{tile_size[1]}).")
                log.info("The current mapproxy configuration is invalid for this geopackage.")
                raise ValueError("tile_size is improperly configured.")
            if not is_close(gpkg_pixel_x_size, resolution) or not is_close(gpkg_pixel_y_size, resolution):
                log.info(
                    f"The geopackage {self.geopackage_file} table name {self.table_name} already exists and level"
                    f" {gpkg_zoom_level} a resolution of ({gpkg_pixel_x_size:.13f},{gpkg_pixel_y_size:.13f})"
                    f" which is different than the configured resolution of ({resolution:.13f},{resolution:.13f}).")
                log.info("The current mapproxy configuration is invalid for this geopackage.")
                raise ValueError("res is improperly configured.")
            return True

    def _initialize_gpkg(self):
        log.info('initializing Geopackage file %s', self.geopackage_file)
        with sqlite3.connect(self.geopackage_file, timeout=self.timeout) as db:

            if self.wal:
                db.execute('PRAGMA journal_mode=wal')

            proj = get_epsg_num(self.tile_grid.srs.srs_code)

            db.execute(create_gpkg_contents_statement)
            db.execute(create_spatial_ref_sys_statment)
            db.execute(create_tile_matrix_statement)
            db.execute(create_tile_matrix_set_statement)
            db.execute(create_table_statement.format(self.table_name))

            db.execute("PRAGMA foreign_keys = 1;")

            # List of WKT execute statements and data.("""
            wkt_statement = """
                INSERT OR REPLACE INTO gpkg_spatial_ref_sys (
                    srs_id,
                    organization,
                    organization_coordsys_id,
                    srs_name,
                    definition)
                VALUES (?, ?, ?, ?, ?)
            """
            wkt_entries = [
                (3857, 'epsg', 3857, 'WGS 84 / Pseudo-Mercator', proj_string_3857),
                (4326, 'epsg', 4326, 'WGS 84', proj_string_4326),
                (-1, 'NONE', -1, ' ', 'undefined'),
                (0, 'NONE', 0, ' ', 'undefined')
            ]

            if get_epsg_num(self.tile_grid.srs.srs_code) not in [4326, 3857]:
                wkt_entries.append((proj, 'epsg', proj, 'Not provided', "Added via Mapproxy."))
            db.commit()

            # Add geopackage version to the header (1.0)
            db.execute("PRAGMA application_id = 1196437808;")
            db.commit()

            for wkt_entry in wkt_entries:
                try:
                    db.execute(wkt_statement, wkt_entry)
                except sqlite3.IntegrityError:
                    log.info("srs_id already exists.")
            db.commit()

            last_change = datetime.fromtimestamp(
                int(os.environ.get('SOURCE_DATE_EPOCH', time.time())),
                timezone.utc
            )

            # Ensure that tile table exists here, don't overwrite a valid entry.
            try:
                db.execute("""
                    INSERT INTO gpkg_contents (
                        table_name,
                        data_type,
                        identifier,
                        description,
                        last_change,
                        min_x,
                        max_x,
                        min_y,
                        max_y,
                        srs_id)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?);
                    """, (self.table_name,
                          "tiles",
                          self.table_name,
                          "Created with Mapproxy.",
                          last_change,
                          self.bbox[0],
                          self.bbox[2],
                          self.bbox[1],
                          self.bbox[3],
                          proj))
            except sqlite3.IntegrityError:
                pass
            db.commit()

            # Ensure that tile set exists here, don't overwrite a valid entry.
            try:
                db.execute("""
                    INSERT INTO gpkg_tile_matrix_set (table_name, srs_id, min_x, max_x, min_y, max_y)
                    VALUES (?, ?, ?, ?, ?, ?);
                """, (
                    self.table_name, proj, self.bbox[0], self.bbox[2], self.bbox[1], self.bbox[3]))
            except sqlite3.IntegrityError:
                pass
            db.commit()

            tile_size = self.tile_grid.tile_size
            for grid, resolution, level in zip(self.tile_grid.grid_sizes,
                                               self.tile_grid.resolutions, range(20)):
                db.execute("""
                    INSERT OR REPLACE INTO gpkg_tile_matrix (table_name, zoom_level, matrix_width,
                        matrix_height, tile_width, tile_height, pixel_x_size, pixel_y_size)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?);
                    """, (self.table_name, level, grid[0], grid[1], tile_size[0], tile_size[1], resolution, resolution))
            db.commit()

        if self.file_permissions is not None:
            permission = int(self.file_permissions, base=8)
            log.info("setting file permissions on GeoPackage: ", permission)
            os.chmod(self.geopackage_file, permission)

    def is_cached(self, tile, dimensions=None):
        if tile.coord is None:
            return True
        if tile.source:
            return True

        return self.load_tile(tile, dimensions=dimensions)

    def store_tile(self, tile, dimensions=None):
        if tile.stored:
            return True
        return self._store_bulk([tile])

    def store_tiles(self, tiles, dimensions=None):
        tiles = [t for t in tiles if not t.stored]
        return self._store_bulk(tiles)

    def _store_bulk(self, tiles):
        records = []
        # tile_buffer (as_buffer) will encode the tile to the target format
        # we collect all tiles before, to avoid having the db transaction
        # open during this slow encoding
        for tile in tiles:
            with tile_buffer(tile) as buf:
                content = buf.read()
                x, y, level = tile.coord
                records.append((level, x, y, content))

        cursor = self.db.cursor()
        try:
            stmt = (f"INSERT OR REPLACE INTO [{self.table_name}] (zoom_level, tile_column, tile_row, tile_data)"
                    f" VALUES (?,?,?,?)")
            cursor.executemany(stmt, records)
            self.db.commit()
        except sqlite3.OperationalError as ex:
            log.warning('unable to store tile: %s', ex)
            return False
        return True

    def load_tile(self, tile, with_metadata=False, dimensions=None):
        if tile.source or tile.coord is None:
            return True

        cur = self.db.cursor()
        cur.execute("""SELECT tile_data FROM [{0}]
                WHERE tile_column = ? AND
                      tile_row = ? AND
                      zoom_level = ?""".format(self.table_name), tile.coord)

        content = cur.fetchone()
        if content:
            tile.source = ImageSource(BytesIO(content[0]))
            return True
        else:
            return False

    def load_tiles(self, tiles, with_metadata=False, dimensions=None):
        # associate the right tiles with the cursor
        tile_dict = {}
        coords = []
        for tile in tiles:
            if tile.source or tile.coord is None:
                continue
            x, y, level = tile.coord
            coords.append(x)
            coords.append(y)
            coords.append(level)
            tile_dict[(x, y)] = tile

        if not tile_dict:
            # all tiles loaded or coords are None
            return True

        stmt_base = "SELECT tile_column, tile_row, tile_data FROM [{0}] WHERE ".format(self.table_name)

        loaded_tiles = 0

        # SQLite is limited to 1000 args -> split into multiple requests if more arguments are needed
        while coords:
            cur_coords = coords[:999]

            stmt = stmt_base + ' OR '.join(
                ['(tile_column = ? AND tile_row = ? AND zoom_level = ?)'] * (len(cur_coords) // 3))

            cursor = self.db.cursor()
            cursor.execute(stmt, cur_coords)

            for row in cursor:
                loaded_tiles += 1
                tile = tile_dict[(row[0], row[1])]
                data = row[2]
                tile.size = len(data)
                tile.source = ImageSource(BytesIO(data))
            cursor.close()

            coords = coords[999:]

        return loaded_tiles == len(tile_dict)

    def remove_tile(self, tile, dimensions=None):
        cursor = self.db.cursor()
        cursor.execute(
            "DELETE FROM [{0}] WHERE (tile_column = ? AND tile_row = ? AND zoom_level = ?)".format(self.table_name),
            tile.coord)
        self.db.commit()
        if cursor.rowcount:
            return True
        return False

    def remove_level_tiles_before(self, level, timestamp=None, remove_all=False):
        if remove_all:
            cursor = self.db.cursor()
            cursor.execute(
                "DELETE FROM [{0}] WHERE (zoom_level = ?)".format(self.table_name), (level,))
            self.db.commit()
            log.info("Cursor rowcount = {0}".format(cursor.rowcount))
            if cursor.rowcount:
                return True
            return False

    def load_tile_metadata(self, tile, dimensions=None):
        if not self.supports_timestamp:
            # GPKG specification does not include tile timestamps.
            # This sets the timestamp of the tile to epoch (1970s)
            tile.timestamp = -1
        else:
            self.load_tile(tile, dimensions=dimensions)


class GeopackageLevelCache(TileCacheBase):

    def __init__(self, geopackage_dir, tile_grid, table_name, timeout=30, wal=False, coverage=None,
                 directory_permissions=None, file_permissions=None):
        super(GeopackageLevelCache, self).__init__(coverage)
        md5 = hashlib.new('md5', geopackage_dir.encode('utf-8'), usedforsecurity=False)
        self.lock_cache_id = 'gpkg-' + md5.hexdigest()
        self.cache_dir = geopackage_dir
        self.tile_grid = tile_grid
        self.table_name = table_name
        self.timeout = timeout
        self.wal = wal
        self._geopackage = {}
        self._geopackage_lock = threading.Lock()
        self.directory_permissions = directory_permissions
        self.file_premissions = file_permissions

    def _get_level(self, level):
        if level in self._geopackage:
            return self._geopackage[level]

        with self._geopackage_lock:
            if level not in self._geopackage:
                geopackage_filename = os.path.join(self.cache_dir, '%s.gpkg' % level)
                self._geopackage[level] = GeopackageCache(
                    geopackage_filename,
                    self.tile_grid,
                    self.table_name,
                    with_timestamps=False,
                    timeout=self.timeout,
                    wal=self.wal,
                    coverage=self.coverage,
                    directory_permissions=self.directory_permissions,
                    file_permissions=self.file_premissions
                )

        return self._geopackage[level]

    def cleanup(self):
        """
        Close all open connection and remove them from cache.
        """
        with self._geopackage_lock:
            for gp in self._geopackage.values():
                gp.cleanup()

    def is_cached(self, tile, dimensions=None):
        if tile.coord is None:
            return True
        if tile.source:
            return True

        return self._get_level(tile.coord[2]).is_cached(tile, dimensions=dimensions)

    def store_tile(self, tile, dimensions=None):
        if tile.stored:
            return True

        return self._get_level(tile.coord[2]).store_tile(tile, dimensions=dimensions)

    def store_tiles(self, tiles, dimensions=None):
        failed = False
        for level, tiles in groupby(tiles, key=lambda t: t.coord[2]):
            tiles = [t for t in tiles if not t.stored]
            res = self._get_level(level).store_tiles(tiles, dimensions=dimensions)
            if not res:
                failed = True
        return failed

    def load_tile(self, tile, with_metadata=False, dimensions=None):
        if tile.source or tile.coord is None:
            return True

        return self._get_level(tile.coord[2]).load_tile(tile, with_metadata=with_metadata, dimensions=dimensions)

    def load_tiles(self, tiles, with_metadata=False, dimensions=None):
        level = None
        for tile in tiles:
            if tile.source or tile.coord is None:
                continue
            level = tile.coord[2]
            break

        if not level:
            return True

        return self._get_level(level).load_tiles(tiles, with_metadata=with_metadata, dimensions=dimensions)

    def remove_tile(self, tile, dimensions=None):
        if tile.coord is None:
            return True

        return self._get_level(tile.coord[2]).remove_tile(tile, dimensions=dimensions)

    def remove_level_tiles_before(self, level, timestamp=None, remove_all=False):
        level_cache = self._get_level(level)
        if remove_all:
            level_cache.cleanup()
            os.unlink(level_cache.geopackage_file)
            return True
        else:
            return level_cache.remove_level_tiles_before(level, timestamp)

    def load_tile_metadata(self, tile, dimensions=None):
        return self._get_level(tile.coord[2]).load_tile_metadata(tile, dimensions=dimensions)


def is_close(a, b, rel_tol=1e-09, abs_tol=0.0):
    """
    See PEP 485, added here for legacy versions.

    >>> is_close(0.0, 0.0)
    True
    >>> is_close(1, 1.0)
    True
    >>> is_close(0.01, 0.001)
    False
    >>> is_close(0.0001001, 0.0001, rel_tol=1e-02)
    True
    >>> is_close(0.0001001, 0.0001)
    False

    @param a: An int or float.
    @param b: An int or float.
    @param rel_tol: Relative tolerance - maximumed allow difference between two numbers.
    @param abs_tol: Absolute tolerance - minimum absolute tolerance.
    @return: True if the values a and b are close.

    """
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


create_gpkg_contents_statement = """
    CREATE TABLE IF NOT EXISTS gpkg_contents(
        table_name  TEXT     NOT NULL PRIMARY KEY,
            -- The name of the tiles, or feature table
        data_type   TEXT     NOT NULL,
            -- Type of data stored in the table: "features" per clause Features
            -- (http://www.geopackage.org/spec/#features), "tiles" per clause Tiles
            -- (http://www.geopackage.org/spec/#tiles), or an implementer-defined value for other data
            -- tables per clause in an Extended GeoPackage
        identifier  TEXT     UNIQUE,
            -- A human-readable identifier (e.g. short name) for the table_name content
        description TEXT     DEFAULT '',
            -- A human-readable description for the table_name content
        last_change DATETIME NOT NULL DEFAULT (strftime('%Y-%m-%dT%H:%M:%fZ','now')),
            -- Timestamp value in ISO 8601 format as defined by the strftime function %Y-%m-%dT%H:%M:%fZ
            -- format string applied to the current time
        min_x       DOUBLE,
            -- Bounding box minimum easting or longitude for all content in table_name
        min_y       DOUBLE,
            -- Bounding box minimum northing or latitude for all content in table_name
        max_x       DOUBLE,
            -- Bounding box maximum easting or longitude for all content in table_name
        max_y       DOUBLE,
            -- Bounding box maximum northing or latitude for all content in table_name
        srs_id      INTEGER,
            -- Spatial Reference System ID: gpkg_spatial_ref_sys.srs_id; when data_type is features,
            -- SHALL also match gpkg_geometry_columns.srs_id; When data_type is tiles, SHALL also match
            -- gpkg_tile_matrix_set.srs.id
        CONSTRAINT fk_gc_r_srs_id FOREIGN KEY (srs_id) REFERENCES gpkg_spatial_ref_sys(srs_id))
"""

create_spatial_ref_sys_statment = """
    CREATE TABLE IF NOT EXISTS gpkg_spatial_ref_sys(
        srs_name                 TEXT    NOT NULL,
          -- Human readable name of this SRS (Spatial Reference System)
        srs_id                   INTEGER NOT NULL PRIMARY KEY,
          -- Unique identifier for each Spatial Reference System within a GeoPackage
        organization             TEXT    NOT NULL,
          -- Case-insensitive name of the defining organization e.g. EPSG or epsg
        organization_coordsys_id INTEGER NOT NULL,
          -- Numeric ID of the Spatial Reference System assigned by the organization
        definition               TEXT    NOT NULL,
          -- Well-known Text representation of the Spatial Reference System
        description              TEXT)
"""

create_tile_matrix_statement = """
    CREATE TABLE IF NOT EXISTS gpkg_tile_matrix
         (table_name    TEXT    NOT NULL, -- Tile Pyramid User Data Table Name
          zoom_level    INTEGER NOT NULL, -- 0 <= zoom_level <= max_level for table_name
          matrix_width  INTEGER NOT NULL, -- Number of columns (>= 1) in tile matrix at this zoom level
          matrix_height INTEGER NOT NULL, -- Number of rows (>= 1) in tile matrix at this zoom level
          tile_width    INTEGER NOT NULL, -- Tile width in pixels (>= 1) for this zoom level
          tile_height   INTEGER NOT NULL, -- Tile height in pixels (>= 1) for this zoom level
          pixel_x_size  DOUBLE  NOT NULL, -- In t_table_name srid units or default meters for srid 0 (>0)
          pixel_y_size  DOUBLE  NOT NULL, -- In t_table_name srid units or default meters for srid 0 (>0)
          CONSTRAINT pk_ttm PRIMARY KEY (table_name, zoom_level),
          CONSTRAINT fk_tmm_table_name FOREIGN KEY (table_name) REFERENCES gpkg_contents(table_name))
"""

create_tile_matrix_set_statement = """
    CREATE TABLE IF NOT EXISTS gpkg_tile_matrix_set(
        table_name TEXT    NOT NULL PRIMARY KEY,
            -- Tile Pyramid User Data Table Name
        srs_id     INTEGER NOT NULL,
            -- Spatial Reference System ID: gpkg_spatial_ref_sys.srs_id
        min_x      DOUBLE  NOT NULL,
            -- Bounding box minimum easting or longitude for all content in table_name
        min_y      DOUBLE  NOT NULL,
            -- Bounding box minimum northing or latitude for all content in table_name
        max_x      DOUBLE  NOT NULL,
            -- Bounding box maximum easting or longitude for all content in table_name
        max_y      DOUBLE  NOT NULL,
            -- Bounding box maximum northing or latitude for all content in table_name
        CONSTRAINT fk_gtms_table_name FOREIGN KEY (table_name) REFERENCES gpkg_contents(table_name),
        CONSTRAINT fk_gtms_srs FOREIGN KEY (srs_id) REFERENCES gpkg_spatial_ref_sys (srs_id))
"""

create_table_statement = """
    CREATE TABLE IF NOT EXISTS [{0}]
        (id          INTEGER PRIMARY KEY AUTOINCREMENT, -- Autoincrement primary key
         zoom_level  INTEGER NOT NULL,                  -- min(zoom_level) <= zoom_level <= max(zoom_level)
            -- for t_table_name
         tile_column INTEGER NOT NULL,                  -- 0 to tile_matrix matrix_width - 1
         tile_row    INTEGER NOT NULL,                  -- 0 to tile_matrix matrix_height - 1
         tile_data   BLOB    NOT NULL,                  -- Of an image MIME type specified in clauses Tile
            -- Encoding PNG, Tile Encoding JPEG, Tile Encoding WEBP
         UNIQUE (zoom_level, tile_column, tile_row))
"""

proj_string_3857 = """
    PROJCS["WGS 84 / Pseudo-Mercator",GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,\
    AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],\
    UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]],\
    PROJECTION["Mercator_1SP"],PARAMETER["central_meridian",0],PARAMETER["scale_factor",1],PARAMETER["false_easting",0],\
    PARAMETER["false_northing",0],UNIT["metre",1,AUTHORITY["EPSG","9001"]],AXIS["X",EAST],AXIS["Y",NORTH],\
    AUTHORITY["EPSG","3857"]]\
"""

proj_string_4326 = """
    GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],\
    AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,\
    AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]\
"""



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mapproxy.config.config import base_config
from mapproxy.grid.resolutions import res_to_ogc_scale
from mapproxy.request.base import Request
from mapproxy.service.ogcapi.constants import (
    FORMAT_TYPES,
    F_JSON,
    F_HTML,
    F_PNG,
    F_JPEG,
)
from mapproxy.service.ogcapi.server import OGCAPIServer
from mapproxy.service.wms import WMSLayerBase
from mapproxy.srs import SRS


def _get(d: dict, *path):
    for p in path:
        if p not in d:
            return None
        d = d[p]
    return d


def get_collection(server: OGCAPIServer, req: Request, layer: WMSLayerBase):
    id = layer.name
    col = {}
    col["dataType"] = "map"
    col["id"] = id
    col["title"] = layer.title
    abstract = _get(layer.md, "abstract")
    if abstract:
        col["description"] = abstract

    attribution = layer.md.get("attribution", None)
    if attribution:
        attribution_title = attribution.get("title", None)
        if attribution_title:
            col["attribution"] = attribution_title
            attribution_url = attribution.get("url", None)
            if attribution_url:
                attribution_logo_url = _get(attribution, "logo", "url")
                if attribution_logo_url:
                    col[
                        "attribution"
                    ] = f"[![{attribution_title}]({attribution_logo_url}]({attribution_url})"
                else:
                    col["attribution"] = f"[{attribution_title}]({attribution_url})"
                col["attributionMediaType"] = "text/markdown"

    #  The OGC API Common specfication required that the
    # ``"extent": { "spatial": { "bbox": [ [xmin,ymin,xmax,ymal] ] }`` property
    # is always presented in CRS84 (or for planetary datasets in the applicable
    # global geographic CRS, that you then set as a ``crs`` child member of crs.
    if layer.extent.srs.srs_code.startswith("EPSG:"):
        bbox_crs = SRS(4326)
    else:
        # Non-Earth data
        bbox_crs = layer.extent.srs.get_geographic_srs()

    bbox = layer.extent.transform(bbox_crs).bbox
    col["extent"] = {"spatial": {"bbox": [bbox]}}
    if bbox_crs.srs_code != "EPSG:4326":
        col["extent"]["spatial"]["crs"] = bbox_crs.to_ogc_url()

    # When the native bbox is not in CRS84/EPSG:4326, you can add an optional
    # "extent"["spatial"]["storageCrsBbox"] member to express it also in its #
    # native CRS.
    if layer.extent.srs != bbox_crs:
        col["extent"]["spatial"]["storageCrsBbox"] = [layer.extent.bbox]

    crs = []
    if layer.extent.srs.srs_code.startswith("EPSG:"):
        crs.append(SRS("OGC:CRS84").to_ogc_url())
        crs.append(SRS(4326).to_ogc_url())
    for srs in server.map_srs:
        if srs.srs_code != "EPSG:4326":
            crs.append(srs.to_ogc_url())
    col["crs"] = crs

    storageSRS = layer.extent.srs
    if storageSRS.srs_code == "EPSG:4326":
        col["storageCrs"] = SRS("OGC:CRS84").to_ogc_url()
    else:
        col["storageCrs"] = storageSRS.to_ogc_url()

    is_html = server.is_html_req(req)
    col["links"] = [
        {
            "rel": ("self" if not is_html else "alternate"),
            "type": FORMAT_TYPES[F_JSON],
            "title": "The JSON representation of this data collection",
            "href": server.create_href(req, f"/ogcapi/collections/{id}?f={F_JSON}"),
        },
        {
            "rel": ("alternate" if not is_html else "self"),
            "type": FORMAT_TYPES[F_HTML],
            "title": "The HTML representation of this data collection",
            "href": server.create_href(req, f"/ogcapi/collections/{id}?f={F_HTML}"),
        },
    ]
    if server.enable_maps:
        col["links"] += [
            {
                "rel": "http://www.opengis.net/def/rel/ogc/1.0/map",
                "type": FORMAT_TYPES[F_PNG],
                "title": "Default map (as PNG)",
                "href": server.create_href(req, f"/ogcapi/collections/{id}/map.png"),
            },
            {
                "rel": "http://www.opengis.net/def/rel/ogc/1.0/map",
                "type": FORMAT_TYPES[F_JPEG],
                "title": "Default map (as JPEG)",
                "href": server.create_href(req, f"/ogcapi/collections/{id}/map.jpg"),
            },
        ]
    if server.enable_tiles:
        col["links"] += [
            {
                "rel": "http://www.opengis.net/def/rel/ogc/1.0/tilesets-map",
                "type": FORMAT_TYPES[F_JSON],
                "title": "Map tilesets available for this collection (as JSON)",
                "href": server.create_href(
                    req, f"/ogcapi/collections/{id}/map/tiles?f={F_JSON}"
                ),
            },
            {
                "rel": "http://www.opengis.net/def/rel/ogc/1.0/tilesets-map",
                "type": FORMAT_TYPES[F_HTML],
                "title": "Map tilesets available for this collection (as HTML)",
                "href": server.create_href(
                    req, f"/ogcapi/collections/{id}/map/tiles?f={F_HTML}"
                ),
            },
        ]

    if layer.res_range and layer.res_range.max_res:
        # Not a typo. If my understanding is correct, MapProxy max_res
        # corresponds to OGC minCellSize
        col["minCellSize"] = layer.res_range.max_res
        col["minScaleDenominator"] = res_to_ogc_scale(col["minCellSize"])
    if layer.res_range and layer.res_range.min_res:
        # Not a typo. If my understanding is correct, MapProxy min_res
        # corresponds to OGC maxCellSize
        col["maxCellSize"] = layer.res_range.min_res
        col["maxScaleDenominator"] = res_to_ogc_scale(col["maxCellSize"])

    return col


def collections(server: OGCAPIServer, req: Request):
    log = server.log
    log.debug("Collections page")

    for arg in req.args:
        if arg != "f":
            raise OGCAPIServer.unknown_query_parameter(arg)
    if req.args.get("f", None) not in (None, F_JSON, F_HTML):
        raise OGCAPIServer.invalid_parameter("Invalid value for f query parameter")

    is_html = server.is_html_req(req)
    collections = [
        get_collection(server, req, server.layers[id]) for id in server.layers
    ]
    json_resp = {"collections": collections}

    json_resp["links"] = [
        {
            "rel": ("self" if not is_html else "alternate"),
            "type": FORMAT_TYPES[F_JSON],
            "title": "The JSON representation of the list of all data collections served from this endpoint",
            "href": server.create_href(req, f"/ogcapi/collections?f={F_JSON}"),
        },
        {
            "rel": ("alternate" if not is_html else "self"),
            "type": FORMAT_TYPES[F_HTML],
            "title": "The HTML representation of the list of all data collections served from this endpoint",
            "href": server.create_href(req, f"/ogcapi/collections?f={F_HTML}"),
        },
    ]

    if is_html:
        json_resp["collections_path"] = server.create_href(req, "/ogcapi/collections")

    return server.create_json_or_html_response(req, json_resp, "collections/index.html")


def collection(server: OGCAPIServer, req: Request, coll_id: str):
    log = server.log
    log.debug(f"Collection page for {coll_id}")

    for arg in req.args:
        if arg != "f":
            raise OGCAPIServer.unknown_query_parameter(arg)
    if req.args.get("f", None) not in (None, F_JSON, F_HTML):
        raise OGCAPIServer.invalid_parameter("Invalid value for f query parameter")

    if coll_id not in server.layers:
        raise OGCAPIServer.collection_not_found()

    layer = server.layers[coll_id]
    json_resp = get_collection(server, req, layer)

    is_html = server.is_html_req(req)
    if is_html:
        json_resp["collections_path"] = server.create_href(req, "/ogcapi/collections")
        json_resp["srs"] = layer.extent.srs.srs_code
        json_resp["extent"] = list(layer.extent.bbox)
        json_resp["background_url"] = base_config().background.url

    return server.create_json_or_html_response(
        req, json_resp, "collections/collection.html"
    )



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from mapproxy.util.ogr import OGRShapeReader, libgdal


polygon_shapefile = os.path.join(os.path.dirname(__file__), "polygons", "polygons.shp")
polygon_geojson = os.path.join(os.path.dirname(__file__), "polygons", "polygons.geojson")


@pytest.mark.skipif(not libgdal, reason="libgdal not found")
class TestOGRShapeReader(object):

    @pytest.fixture
    def readers(self):
        return [OGRShapeReader(polygon_shapefile), OGRShapeReader(polygon_geojson)]

    def test_read_all(self, readers):
        for reader in readers:
            wkts = list(reader.wkts())
            assert len(wkts) == 3
            for wkt in wkts:
                assert wkt.startswith(b"POLYGON ("), "unexpected WKT: %s" % wkt

    def test_read_filter(self, readers):
        for reader in readers:
            wkts = list(reader.wkts(where="name = 'germany'"))
            assert len(wkts) == 2
            for wkt in wkts:
                assert wkt.startswith(b"POLYGON ("), "unexpected WKT: %s" % wkt

    def test_read_filter_no_match(self, readers):
        for reader in readers:
            wkts = list(reader.wkts(where="name = 'foo'"))
            assert len(wkts) == 0



# This file is part of the MapProxy project.
# Copyright (C) 2022 Even Rouault
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import pytest

from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest
from webtest import AppError


@pytest.fixture(scope="module")
def config_file():
    return "demo.yaml"


class TestDemo(SysTest):

    def test_basic(self, app):
        resp = app.get("/demo/", status=200)
        assert resp.content_type == "text/html"
        assert 'href="../service?REQUEST=GetCapabilities&SERVICE=WMS"' in resp
        assert 'href="../service?REQUEST=GetCapabilities&SERVICE=WMS&tiled=true"' in resp
        assert 'href="../demo/?wmts_layer=wms_cache&format=jpeg&srs=EPSG%3A900913"' in resp
        assert 'href="../demo/?tms_layer=wms_cache&format=jpeg&srs=EPSG%3A900913"' in resp

    def test_previewmap(self, app):
        resp = app.get("/demo/?srs=EPSG%3A3857&format=image%2Fpng&wms_layer=wms_cache", status=200)
        assert resp.content_type == "text/html"
        assert '<h2>Layer Preview - wms_cache</h2>' in resp

    def test_layers_sorted_by_name(self, app):
        resp = app.get("/demo/", status=200)

        patternTable = re.compile(r'<table class="code">.*?</table>', re.DOTALL)
        patternWMS = r'<td\s+rowspan="2">([\w\d-]+)</td>'
        patternWMTS_TBS = r'<td\s+rowspan="1">([\w\d-]+)</td>'
        tables = patternTable.findall(resp.text)

        layersWMS = re.findall(patternWMS, resp.text, re.IGNORECASE)
        layersWMTS = re.findall(patternWMTS_TBS, tables[1], re.IGNORECASE)
        layersTBS = re.findall(patternWMTS_TBS, tables[2], re.IGNORECASE)

        assert layersWMS == sorted(layersWMS)
        assert layersWMTS == sorted(layersWMTS)
        assert layersTBS == sorted(layersTBS)

    def test_external(self, app):
        expected_req = (
            {
                "path": r"/path/service?REQUEST=GetCapabilities&SERVICE=WMS"
            },
            {"body": b"test-string", "headers": {"content-type": "text/xml"}}
        )
        with mock_httpd(
                ("localhost", 42423), [expected_req]
        ):
            resp = app.get('/demo/?wms_capabilities&type=external', extra_environ={
                'HTTP_X_FORWARDED_HOST': 'localhost:42423/path'
            })
            content = resp.text
            assert 'test-string' in content
            assert 'http://localhost:42423/path/service?REQUEST=GetCapabilities&SERVICE=WMS' in content

    def test_external_xss_injection(self, app):
        expected_req = (
            {
                "path": r"/path/&gt;&lt;script&gt;alert(XSS)&lt;/script&gt;/service?REQUEST=GetCapabilities&SERVICE=WMS"
            },
            {"body": b"test-string", "headers": {"content-type": "text/xml"}}
        )

        with mock_httpd(
                ("localhost", 42423), [expected_req]
        ):
            resp = app.get('/demo/?wms_capabilities&type=external', extra_environ={
                'HTTP_X_FORWARDED_HOST': 'localhost:42423/path/"><script>alert(\'XSS\')</script>'
            })
        content = resp.text
        assert 'test-string' in content
        assert '"><script>alert(\'XSS\')' not in content
        assert '&gt;&lt;script&gt;alert(XSS)&lt;/script&gt;' in content

    def test_external_file_protocol(self, app):
        try:
            app.get('/demo/?wms_capabilities&type=external', extra_environ={
                'HTTP_X_FORWARDED_PROTO': 'file'
            })
        except AppError as e:
            assert '400 Bad Request' in e.args[0]

    def test_tms_layer_xss(self, app):
        expected_req = (
            {
                "path": r"/tms/1.0.0/osm&gt;&lt;script&gt;alert(XSS)&lt;/script&gt;/1.0.0"
            },
            {"body": b"test-string", "headers": {"content-type": "text/xml"}}
        )

        with mock_httpd(
                ("localhost", 42423), [expected_req]
        ):
            resp = app.get(
                '/demo/?tms_capabilities&layer=osm"><script>alert(\'XSS\')</script>&type=external&srs=1.0.0',
                extra_environ={
                    'HTTP_X_FORWARDED_HOST': 'localhost:42423'
                }
            )
        content = resp.text
        assert '"><script>alert(\'XSS\')' not in content
        assert '&gt;&lt;script&gt;alert(XSS)&lt;/script&gt;' in content



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Retrieve tiles from different tile servers (TMS/TileCache/etc.).
"""

import sys
from mapproxy.image.opts import ImageOptions
from mapproxy.source import SourceError
from mapproxy.client.http import HTTPClientError
from mapproxy.source import InvalidSourceQuery
from mapproxy.layer import BlankImage, map_extent_from_grid, CacheMapLayer, MapLayer
from mapproxy.util.py import reraise_exception

import logging
log = logging.getLogger('mapproxy.source.tile')
log_config = logging.getLogger('mapproxy.config')


class TiledSource(MapLayer):
    def __init__(self, grid, client, coverage=None, image_opts=None, error_handler=None,
                 res_range=None):
        MapLayer.__init__(self, image_opts=image_opts)
        self.grid = grid
        self.client = client
        self.image_opts = image_opts or ImageOptions()
        self.coverage = coverage
        self.extent = coverage.extent if coverage else map_extent_from_grid(grid)
        self.res_range = res_range
        self.error_handler = error_handler

    def get_map(self, query):
        if self.grid.tile_size != query.size:
            ex = InvalidSourceQuery(
                'tile size of cache and tile source do not match: %s != %s'
                % (self.grid.tile_size, query.size)
            )
            log_config.error(ex)
            raise ex

        if self.grid.srs != query.srs:
            ex = InvalidSourceQuery(
                'SRS of cache and tile source do not match: %r != %r'
                % (self.grid.srs, query.srs)
            )
            log_config.error(ex)
            raise ex

        if self.res_range and not self.res_range.contains(query.bbox, query.size,
                                                          query.srs):
            raise BlankImage()
        if self.coverage and not self.coverage.intersects(query.bbox, query.srs):
            raise BlankImage()

        _bbox, grid, tiles = self.grid.get_affected_tiles(query.bbox, query.size)

        if grid != (1, 1):
            raise InvalidSourceQuery('BBOX does not align to tile')

        tile_coord = next(tiles)

        try:
            return self.client.get_tile(tile_coord, format=query.format)
        except HTTPClientError as e:
            if self.error_handler:
                resp = self.error_handler.handle(e.response_code, query)
                if resp:
                    return resp
            log.warning('could not retrieve tile (client: %s): %s', self.client, e, exc_info=True)
            reraise_exception(SourceError(e.args[0]), sys.exc_info())


class CacheSource(CacheMapLayer):
    def __init__(self, tile_manager, extent=None, image_opts=None,
                 max_tile_limit=None, tiled_only=False):
        CacheMapLayer.__init__(self, tile_manager, extent=extent, image_opts=image_opts,
                               max_tile_limit=max_tile_limit)
        self.supports_meta_tiles = not tiled_only
        self.tiled_only = tiled_only

    def get_map(self, query):
        if self.tiled_only:
            query.tiled_only = True
        return CacheMapLayer.get_map(self, query)



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mapproxy.request.base import Request
from mapproxy.grid.tile_grid import tile_grid_to_ogc_tile_matrix_set
from mapproxy.service.ogcapi.server import OGCAPIException, OGCAPIServer
from mapproxy.service.ogcapi.constants import FORMAT_TYPES, F_JSON, F_HTML


def tilematrixsets(server: OGCAPIServer, req: Request):
    log = server.log
    log.debug("TileMatrixSets")

    for arg in req.args:
        if arg != "f":
            raise OGCAPIServer.unknown_query_parameter(arg)
    if req.args.get("f", None) not in (None, F_JSON, F_HTML):
        raise OGCAPIServer.invalid_parameter("Invalid value for f query parameter")

    tileMatrixSets = []

    is_html = server.is_html_req(req)

    for name in server.grid_configs:
        tms = {
            "id": name,
            "title": name,
            "links": [
                {
                    "rel": ("self" if not is_html else "alternate"),
                    "type": FORMAT_TYPES[F_JSON],
                    "title": f"The JSON representation of the {name} tiling scheme definition",
                    "href": server.create_href(
                        req, f"/ogcapi/tileMatrixSets/{name}?f={F_JSON}"
                    ),
                },
                {
                    "rel": ("alternate" if not is_html else "self"),
                    "type": FORMAT_TYPES[F_HTML],
                    "title": f"The HTML representation of the {name} tiling scheme definition",
                    "href": server.create_href(
                        req, f"/ogcapi/tileMatrixSets/{name}?f={F_HTML}"
                    ),
                },
            ],
        }
        if name == "WebMercatorQuad":
            tms[
                "uri"
            ] = "http://www.opengis.net/def/tilematrixset/OGC/1.0/WebMercatorQuad"

        tileMatrixSets.append(tms)

    json_resp = {"tileMatrixSets": tileMatrixSets}

    return server.create_json_or_html_response(
        req, json_resp, "tilematrixsets/index.html"
    )


def tilematrixset(server: OGCAPIServer, req: Request, id: str):
    log = server.log
    log.info("TileMatrixSet")

    for arg in req.args:
        if arg != "f":
            raise OGCAPIServer.unknown_query_parameter(arg)
    if req.args.get("f", None) not in (None, F_JSON, F_HTML):
        raise OGCAPIServer.invalid_parameter("Invalid value for f query parameter")

    if id not in server.grid_configs:
        raise OGCAPIException("Not Found", "TileMatrixSet not found", status=404)

    uri = None
    wellKnownScaleSet = None
    if id == "WebMercatorQuad":
        uri = "http://www.opengis.net/def/tilematrixset/OGC/1.0/WebMercatorQuad"
        wellKnownScaleSet = (
            "http://www.opengis.net/def/wkss/OGC/1.0/GoogleMapsCompatible"
        )

    json_resp = tile_grid_to_ogc_tile_matrix_set(
        server.grid_configs[id].tile_grid(),
        id,
        title=id,
        uri=uri,
        wellKnownScaleSet=wellKnownScaleSet,
    )

    return server.create_json_or_html_response(
        req, json_resp, "tilematrixsets/tilematrixset.html"
    )



import yaml
import os

from mapproxy.util.sqlite3 import sqlite3


def conf_from_geopackage(geopackage_file, output_file=None):
    conf = get_geopackage_configuration_dict(geopackage_file)
    yaml.SafeDumper.add_representer(
        type(None),
        lambda dumper, value: dumper.represent_scalar(u'tag:yaml.org,2002:null', '')
    )

    if output_file:
        with open('data.yml', 'w') as outfile:
            outfile.write(yaml.safe_dump(conf, default_flow_style=False))
    else:
        print(yaml.safe_dump(get_geopackage_configuration_dict(geopackage_file), default_flow_style=False))


def get_gpkg_contents(geopackage_file, data_type='tiles'):
    """
    :param geopackage_file: Path to the geopackage file.
    :param data_type: The type of layer to return tiles or features.
    :return: One or more tuples with the table_name, min_x, min_y, max_x, max_y, srs_id for each layer in the
        geopackage.
    """
    with sqlite3.connect(geopackage_file) as db:
        cur = db.execute(
            "SELECT table_name, data_type, identifier, description, last_change, min_x, min_y, max_x, "
            "max_y, srs_id FROM gpkg_contents WHERE data_type = ?", (data_type,))
    return cur.fetchall()


def get_table_organization_coordsys_id(geopackage_file, srs_id):
    """
    :param geopackage_file: Path to the geopackage file.
    :param srs_id: The srs_id which is the key value in the organization_coordsys_id.
    :return: An integer representing the organization_coordsys_id as an EPSG code.
    """
    with sqlite3.connect(geopackage_file) as db:
        cur = db.execute("SELECT organization_coordsys_id FROM gpkg_spatial_ref_sys WHERE srs_id = ?", (srs_id,))
    results = cur.fetchone()
    if results:
        return results[0]


def get_table_tile_matrix(geopackage_file, table_name):
    """
    :param geopackage_file: Path to the geopackage file.
    :param table_name: The table_name associated with the tile_matrix data.
    :return: A tuple of tuple containing zoom_level, matrix_width, matrix_height, tile_width, tile_height, pixel_x_size,
    pixel_y_size for each zoom_level.
    """

    with sqlite3.connect(geopackage_file) as db:
        cur = db.execute("SELECT zoom_level,"
                         "matrix_width, "
                         "matrix_height, "
                         "tile_width, "
                         "tile_height, "
                         "pixel_x_size, "
                         "pixel_y_size "
                         "FROM gpkg_tile_matrix WHERE table_name = ?"
                         "ORDER BY zoom_level", (table_name,))
        return cur.fetchall()


def get_estimated_tile_res_ratio(tile_matrix):
    """

    :param tile_matrix: A tuple of tuples representing the geopackage tile matrix (without the table name included).
    :return: The rate at which the resolution increases between levels.
    """
    default_res_factor = 2
    if len(tile_matrix) < 2:
        return default_res_factor
    layer = tile_matrix[0]
    next_layer = tile_matrix[1]
    return (layer[6]/next_layer[6])/(next_layer[0]-layer[0])


def get_res_table(tile_matrix):
    res_ratio = get_estimated_tile_res_ratio(tile_matrix)
    res_table = []
    if tile_matrix[0][0] == 0:
        first_level_res = tile_matrix[0][5]
    else:
        first_level_res = tile_matrix[0][5]*(res_ratio**tile_matrix[0][0])
    tile_matrix_set = {}
    for level in tile_matrix:
        tile_matrix_set[level[0]] = level
    if not tile_matrix_set.get(0):
        res_table += [first_level_res]
    else:
        res_table += [tile_matrix_set.get(0)[5]]
    for level in range(1, 19):
        res = tile_matrix_set.get(level)
        if not res:
            res_table += [first_level_res/(res_ratio**level)]
        else:
            res_table += [res[5]]
    return res_table


def get_geopackage_configuration_dict(geopackage_file):
    gpkg_contents = get_gpkg_contents(geopackage_file, data_type='tiles')
    conf = {'grids': {},
            'caches': {},
            'layers': [],
            'services': {'demo': None,
                         'tms': {'use_grid_names': True, 'origin': 'nw'},
                         'kml': {'use_grid_names': True},
                         'wmts': None,
                         'wms': None}}

    for gpkg_content in gpkg_contents:
        table_name = str(gpkg_content[0])
        tile_matrix = get_table_tile_matrix(geopackage_file, table_name)
        srs = get_table_organization_coordsys_id(geopackage_file, gpkg_content[9])
        if not tile_matrix or not srs:
            continue
        conf['grids']['{0}_{1}'.format(table_name, srs)] = {'srs': 'EPSG:{0}'.format(srs),
                                                            'tile_size': [tile_matrix[0][3], tile_matrix[0][4]],
                                                            'bbox': [gpkg_content[5],
                                                                     gpkg_content[6],
                                                                     gpkg_content[7],
                                                                     gpkg_content[8]],
                                                            'res': get_res_table(tile_matrix),
                                                            'origin': 'nw'}
        conf['caches']['{0}_cache'.format(table_name)] = {'sources': [],
                                                          'grids': ['{0}_{1}'.format(table_name, srs)],
                                                          'cache': {'type': 'geopackage',
                                                                    'filename': os.path.abspath(geopackage_file),
                                                                    'table_name': table_name}}
        conf['layers'] += [{'name': table_name, 'title': table_name, 'sources': ['{0}_cache'.format(table_name)]}]
    return conf



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import pytest

from mapproxy.test.system import SysTest


class TestWMSBehindProxy(SysTest):
    """
    Check WMS OnlineResources for requests behind HTTP proxies.
    """
    @pytest.fixture(scope='class')
    def config_file(self):
        return 'layer.yaml'

    def test_no_proxy(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities'
                       '&VERSION=1.1.0')
        assert '"http://localhost/service' in resp

    def test_with_script_name(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities'
                       '&VERSION=1.1.0', extra_environ={'HTTP_X_SCRIPT_NAME': '/foo'})
        assert '"http://localhost/service' not in resp
        assert '"http://localhost/foo/service' in resp

    def test_with_host(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities'
                       '&VERSION=1.1.0', extra_environ={'HTTP_HOST': 'example.org'})
        assert '"http://localhost/service' not in resp
        assert '"http://example.org/service' in resp

    def test_with_host_and_script_name(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities'
                       '&VERSION=1.1.0', extra_environ={'HTTP_X_SCRIPT_NAME': '/foo', 'HTTP_HOST': 'example.org'})
        assert '"http://localhost/service' not in resp
        assert '"http://example.org/foo/service' in resp

    def test_with_forwarded_host(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities'
                       '&VERSION=1.1.0', extra_environ={'HTTP_X_FORWARDED_HOST': 'example.org, bar.org'})
        assert '"http://localhost/service' not in resp
        assert '"http://example.org/service' in resp

    def test_with_forwarded_host_and_script_name(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities'
                       '&VERSION=1.1.0',
                       extra_environ={'HTTP_X_FORWARDED_HOST': 'example.org', 'HTTP_X_SCRIPT_NAME': '/foo'})
        assert '"http://localhost/service' not in resp
        assert '"http://example.org/foo/service' in resp

    def test_with_forwarded_proto_and_script_name_and_host(self, app):
        resp = app.get('http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities'
                       '&VERSION=1.1.0', extra_environ={
                           'HTTP_X_FORWARDED_PROTO': 'https',
                           'HTTP_X_SCRIPT_NAME': '/foo',
                           'HTTP_HOST': 'example.org:443'
                       })
        assert '"http://localhost/service' not in resp
        assert '"https://example.org/foo/service' in resp



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from io import BytesIO

import pytest

from mapproxy.request.wms import (
    WMS111MapRequest,
    WMS111FeatureInfoRequest,
    WMS111CapabilitiesRequest,
)
from mapproxy.test.image import is_jpeg
from mapproxy.test.helper import validate_with_dtd
from mapproxy.test.http import assert_no_cache
from mapproxy.test.system.test_wms import is_111_exception
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "layer.yaml"


class TestWMSC(SysTest):

    def setup_method(self):
        self.common_cap_req = WMS111CapabilitiesRequest(
            url="/service?", param=dict(service="WMS", version="1.1.1")
        )
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-20037508,0.0,0.0,20037508",
                width="256",
                height="256",
                layers="wms_cache",
                srs="EPSG:900913",
                format="image/jpeg",
                styles="",
                request="GetMap",
            ),
        )
        self.common_fi_req = WMS111FeatureInfoRequest(
            url="/service?",
            param=dict(
                x="10",
                y="20",
                width="200",
                height="200",
                layers="wms_cache",
                format="image/png",
                query_layers="wms_cache",
                styles="",
                bbox="1000,400,2000,1400",
                srs="EPSG:900913",
            ),
        )

    def test_capabilities(self, app):
        req = str(self.common_cap_req) + "&tiled=true"
        resp = app.get(req)
        xml = resp.lxml
        assert validate_with_dtd(xml, dtd_name="wmsc/1.1.1/WMS_MS_Capabilities.dtd")
        srs = set([e.text for e in xml.xpath("//TileSet/SRS")])
        assert srs == set(["EPSG:4326", "EPSG:900913"])
        assert len(xml.xpath("//TileSet")) == 11

    def test_get_tile(self, app, fixture_cache_data):
        resp = app.get(str(self.common_map_req) + "&tiled=true")
        assert "public" in resp.headers["Cache-Control"]
        assert resp.content_type == "image/jpeg"
        data = BytesIO(resp.body)
        assert is_jpeg(data)

    def test_get_tile_w_rounded_bbox(self, app, fixture_cache_data):
        self.common_map_req.params.bbox = "-20037400,0.0,0.0,20037400"
        resp = app.get(str(self.common_map_req) + "&tiled=true")
        assert "public" in resp.headers["Cache-Control"]
        assert resp.content_type == "image/jpeg"
        data = BytesIO(resp.body)
        assert is_jpeg(data)

    def test_get_tile_wrong_bbox(self, app):
        self.common_map_req.params.bbox = "-20037508,0.0,200000.0,20037508"
        resp = app.get(str(self.common_map_req) + "&tiled=true", expect_errors=True)
        assert_no_cache(resp)
        is_111_exception(resp.lxml, re_msg=".*invalid bbox")

    def test_get_tile_wrong_fromat(self, app):
        self.common_map_req.params.format = "image/png"
        resp = app.get(str(self.common_map_req) + "&tiled=true", expect_errors=True)
        assert_no_cache(resp)
        is_111_exception(resp.lxml, re_msg="Invalid request: invalid.*format.*jpeg")

    def test_get_tile_wrong_size(self, app):
        self.common_map_req.params.size = (256, 255)
        resp = app.get(str(self.common_map_req) + "&tiled=true", expect_errors=True)
        assert_no_cache(resp)
        is_111_exception(resp.lxml, re_msg="Invalid request: invalid.*size.*256x256")



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import optparse
import sys
import yaml

from mapproxy.client.http import HTTPClient
from mapproxy.grid.tile_grid import (
    tile_grid_from_ogc_tile_matrix_set,
    UnsupportedException,
)
from mapproxy.script.conf.utils import MapProxyYAMLDumper
from mapproxy.util.ogcapi import (
    find_href_in_links,
    build_absolute_url,
    normalize_srs_code,
)


def gridconf_from_ogcapitilematrixset_command(args=None):
    parser = optparse.OptionParser(
        "%prog gridconf-from-ogcapitilematrixset url_to_landing_page"
    )
    parser.add_option("--url", dest="url", help="URL to OGC API landing page.")

    from mapproxy.script.util import setup_logging
    import logging

    setup_logging(logging.WARN)

    if args:
        args = args[1:]  # remove script name

    (options, args) = parser.parse_args(args)
    if not options.url:
        if len(args) != 1:
            parser.print_help()
            sys.exit(1)
        else:
            options.url = args[0]

    conf = get_gridconf(options.url)
    if conf:
        print("Configuration to add to mapproxy.yml:")
        print("")
        print(yaml.dump(conf, default_flow_style=False, Dumper=MapProxyYAMLDumper))
    else:
        sys.exit(0)


def get_gridconf(landing_page_url):
    http_client = HTTPClient(landing_page_url)
    headers = {"Accept": "application/json"}
    resp = http_client.open(landing_page_url, headers=headers)
    j = json.loads(resp.read().decode("utf-8"))

    tilematrixsets_href = find_href_in_links(
        j["links"],
        "http://www.opengis.net/def/rel/ogc/1.0/tiling-schemes",
        "application/json",
    )
    if not tilematrixsets_href:
        print("Cannot find a link to tile matrix set definitions")
        return None
    tilematrixsets_url = build_absolute_url(landing_page_url, tilematrixsets_href)

    resp = http_client.open(tilematrixsets_url, headers=headers)
    j = json.loads(resp.read().decode("utf-8"))

    ret = {"grids": {}}

    for tileMatrixSet in j["tileMatrixSets"]:
        tilematrixset_href = find_href_in_links(
            tileMatrixSet["links"],
            "self",
            "application/json",
        )
        if tilematrixset_href:
            tilematrixset_url = build_absolute_url(landing_page_url, tilematrixset_href)
            resp = http_client.open(tilematrixset_url, headers=headers)
            j = json.loads(resp.read().decode("utf-8"))
            try:
                grid = tile_grid_from_ogc_tile_matrix_set(j)
            except UnsupportedException as e:
                print(f"Cannot handle {tilematrixset_url}: {e}", file=sys.stderr)
                continue
            grid_dict = {}
            grid_dict["srs"] = normalize_srs_code(grid.srs.srs_code)
            grid_dict["bbox"] = grid.bbox
            grid_dict["origin"] = grid.origin
            grid_dict["tile_size"] = [x for x in grid.tile_size]
            grid_dict["res"] = [grid.resolutions[level] for level in range(grid.levels)]
            ret["grids"][grid.name] = grid_dict
        else:
            print(f"Cannot find definition for {tilematrixset_href}", file=sys.stderr)

    return ret



# This file is part of the MapProxy project.
# Copyright (C) 2010-2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import multiprocessing
import os
import random
import shutil
import sys
import tempfile
import threading
import time

from mapproxy.util.lock import FileLock, SemLock, cleanup_lockdir, LockTimeout
from mapproxy.util.fs import (
    _force_rename_dir,
    swap_dir,
    cleanup_directory,
    write_atomic,
)
from mapproxy.util.py import reraise_exception
from mapproxy.util.times import timestamp_before
from mapproxy.test.helper import Mocker, assert_permissions

is_win = sys.platform == "win32"


class TestFileLock(Mocker):

    def setup_method(self):
        Mocker.setup_method(self)
        self.lock_dir = tempfile.mkdtemp()
        self.lock_file = os.path.join(self.lock_dir, "lock.lck")

    def teardown_method(self):
        shutil.rmtree(self.lock_dir)
        Mocker.teardown_method(self)

    def test_file_lock_timeout(self):
        lock = self._create_lock()
        assert_locked(self.lock_file)
        lock  # prevent lint warnings

    def test_file_lock(self):
        # Test a lock that becomes free during a waiting lock() call.
        class Lock(threading.Thread):

            def __init__(self, lock_file):
                threading.Thread.__init__(self)
                self.lock_file = lock_file
                self.lock = FileLock(self.lock_file)

            def run(self):
                self.lock.lock()
                time.sleep(0.2)
                self.lock.unlock()

        lock_thread = Lock(self.lock_file)
        start_time = time.time()
        lock_thread.start()

        # wait until thread got the locked
        while not lock_thread.lock._locked:
            time.sleep(0.001)

        # one lock that times out
        assert_locked(self.lock_file)

        # one lock that will get it after some time
        x = FileLock(self.lock_file, timeout=0.3, step=0.001)
        x.lock()

        locked_for = time.time() - start_time
        assert locked_for - 0.2 <= 0.1, "locking took to long?! (rerun if not sure)"

        # cleanup
        x.unlock()
        lock_thread.join()

    def test_lock_cleanup(self):
        old_lock_file = os.path.join(self.lock_dir, "lock_old.lck")
        x = FileLock(old_lock_file)
        x.lock()
        x.unlock()
        mtime = os.stat(old_lock_file).st_mtime
        mtime -= 7 * 60
        os.utime(old_lock_file, (mtime, mtime))

        x = self._create_lock()
        x.unlock()
        assert os.path.exists(old_lock_file)
        assert os.path.exists(self.lock_file)
        cleanup_lockdir(self.lock_dir)

        assert not os.path.exists(old_lock_file)
        assert os.path.exists(self.lock_file)

    def test_concurrent_access(self):
        count_file = os.path.join(self.lock_dir, "count.txt")
        with open(count_file, "wb") as f:
            f.write(b"0")

        def count_up():
            with FileLock(self.lock_file, timeout=60):
                with open(count_file, "r+b") as f:
                    counter = int(f.read().strip())
                    f.seek(0)
                    f.write(str(counter + 1).encode("utf-8"))

        def do_it():
            for x in range(20):
                time.sleep(0.002)
                count_up()

        threads = [threading.Thread(target=do_it) for _ in range(20)]
        [t.start() for t in threads]
        [t.join() for t in threads]

        with open(count_file, "r+b") as f:
            counter = int(f.read().strip())

        assert counter == 400, counter

    def test_remove_on_unlock(self):
        x = FileLock(self.lock_file, remove_on_unlock=True)
        x.lock()
        assert os.path.exists(self.lock_file)
        x.unlock()
        if is_win:  # not removed on windows
            assert os.path.exists(self.lock_file)
        else:
            assert not os.path.exists(self.lock_file)

        if is_win:
            # not possible to remove lock file when lock is held
            pass
        else:
            x.lock()
            assert os.path.exists(self.lock_file)
            os.remove(self.lock_file)
            assert not os.path.exists(self.lock_file)
            # ignore removed lock
            x.unlock()
            assert not os.path.exists(self.lock_file)

    def test_file_lock_permissions(self):
        file_permissions = '775'

        # Test a lock that becomes free during a waiting lock() call.
        class Lock(threading.Thread):

            def __init__(self, lock_file):
                threading.Thread.__init__(self)
                self.lock_file = lock_file
                self.lock = FileLock(self.lock_file, file_permissions=file_permissions)

            def run(self):
                self.lock.lock()
                time.sleep(0.2)
                self.lock.unlock()

        lock_thread = Lock(self.lock_file)
        lock_thread.start()

        # wait until thread got the locked
        while not lock_thread.lock._locked:
            time.sleep(0.001)

        # one lock that times out
        assert_locked(self.lock_file)
        assert_permissions(self.lock_file, file_permissions)

        lock_thread.join()

    def _create_lock(self):
        lock = FileLock(self.lock_file)
        lock.lock()
        return lock


def assert_locked(lock_file, timeout=0.02, step=0.001):
    assert os.path.exists(lock_file)
    x = FileLock(lock_file, timeout=timeout, step=step)
    try:
        x.lock()
        assert False, "file was not locked"
    except LockTimeout:
        pass


class TestSemLock(object):

    def setup_method(self):
        self.lock_dir = tempfile.mkdtemp()
        self.lock_file = os.path.join(self.lock_dir, "lock.lck")

    def teardown_method(self):
        shutil.rmtree(self.lock_dir)

    def count_lockfiles(self):
        return len(glob.glob(self.lock_file + "*"))

    def test_single(self):
        locks = [SemLock(self.lock_file, 1, timeout=0.01) for _ in range(2)]
        locks[0].lock()
        try:
            locks[1].lock()
        except LockTimeout:
            pass
        else:
            assert False, "expected LockTimeout"

    def test_creating(self):
        locks = [SemLock(self.lock_file, 2) for _ in range(3)]

        assert self.count_lockfiles() == 0
        locks[0].lock()
        assert self.count_lockfiles() == 1
        locks[1].lock()
        assert self.count_lockfiles() == 2
        assert os.path.exists(locks[0]._lock._path)
        assert os.path.exists(locks[1]._lock._path)
        locks[0].unlock()
        locks[2].lock()
        locks[2].unlock()
        locks[1].unlock()

    def test_timeout(self):
        locks = [SemLock(self.lock_file, 2, timeout=0.1) for _ in range(3)]

        assert self.count_lockfiles() == 0
        locks[0].lock()
        assert self.count_lockfiles() == 1
        locks[1].lock()
        assert self.count_lockfiles() == 2
        try:
            locks[2].lock()
        except LockTimeout:
            pass
        else:
            assert False, "expected LockTimeout"
        locks[0].unlock()
        locks[2].unlock()

    def test_load(self):
        locks = [SemLock(self.lock_file, 8, timeout=1) for _ in range(20)]

        new_locks = random.sample([x for x in locks if not x._locked], 5)
        for x in new_locks:
            x.lock()

        for _ in range(20):
            old_locks = random.sample([x for x in locks if x._locked], 3)
            for x in old_locks:
                x.unlock()
            assert len([x for x in locks if x._locked]) == 2
            assert len([x for x in locks if not x._locked]) == 18

            new_locks = random.sample([x for x in locks if not x._locked], 3)
            for x in new_locks:
                x.lock()

            assert len([x for x in locks if x._locked]) == 5
            assert len([x for x in locks if not x._locked]) == 15

        assert self.count_lockfiles() == 8


class DirTest(object):

    def setup_method(self):
        self.tmpdir = tempfile.mkdtemp()

    def teardown_method(self):
        if os.path.exists(self.tmpdir):
            shutil.rmtree(self.tmpdir)

    def mkdir(self, name):
        dirname = os.path.join(self.tmpdir, name)
        os.mkdir(dirname)
        self.mkfile(name, dirname=dirname)
        return dirname

    def mkfile(self, name, dirname=None):
        if dirname is None:
            dirname = self.mkdir(name)
        filename = os.path.join(dirname, name + ".txt")
        open(filename, "wb").close()
        return filename


class TestForceRenameDir(DirTest):

    def test_rename(self):
        src_dir = self.mkdir("bar")
        dst_dir = os.path.join(self.tmpdir, "baz")
        _force_rename_dir(src_dir, dst_dir)
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert not os.path.exists(src_dir)

    def test_rename_overwrite(self):
        src_dir = self.mkdir("bar")
        dst_dir = self.mkdir("baz")
        _force_rename_dir(src_dir, dst_dir)
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert not os.path.exists(src_dir)


class TestSwapDir(DirTest):

    def test_swap_dir(self):
        src_dir = self.mkdir("bar")
        dst_dir = os.path.join(self.tmpdir, "baz")

        swap_dir(src_dir, dst_dir)
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert not os.path.exists(src_dir)

    def test_swap_dir_w_old(self):
        src_dir = self.mkdir("bar")
        dst_dir = self.mkdir("baz")

        swap_dir(src_dir, dst_dir)
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert not os.path.exists(src_dir)

    def test_swap_dir_keep_old(self):
        src_dir = self.mkdir("bar")
        dst_dir = self.mkdir("baz")

        swap_dir(src_dir, dst_dir, keep_old=True, backup_ext=".bak")
        assert os.path.exists(dst_dir)
        assert os.path.exists(os.path.join(dst_dir, "bar.txt"))
        assert os.path.exists(dst_dir + ".bak")
        assert os.path.exists(os.path.join(dst_dir + ".bak", "baz.txt"))


class TestCleanupDirectory(DirTest):

    def test_no_remove(self):
        dirs = [self.mkdir("dir" + str(n)) for n in range(10)]
        for d in dirs:
            assert os.path.exists(d), d
        cleanup_directory(self.tmpdir, timestamp_before(minutes=1))
        for d in dirs:
            assert os.path.exists(d), d

    def test_file_handler(self):
        files = []
        file_handler_calls = []

        def file_handler(filename):
            file_handler_calls.append(filename)

        new_date = timestamp_before(weeks=1)
        for n in range(10):
            fname = "foo" + str(n)
            filename = self.mkfile(fname)
            os.utime(filename, (new_date, new_date))
            files.append(filename)

        for filename in files:
            assert os.path.exists(filename), filename
        cleanup_directory(self.tmpdir, timestamp_before(), file_handler=file_handler)
        for filename in files:
            assert os.path.exists(filename), filename

        assert set(files) == set(file_handler_calls)

    def test_no_directory(self):
        cleanup_directory(os.path.join(self.tmpdir, "invalid"), timestamp_before())
        # nothing should happen

    def test_remove_all(self):
        files = []
        new_date = timestamp_before(weeks=1)
        for n in range(10):
            fname = "foo" + str(n)
            filename = self.mkfile(fname)
            os.utime(filename, (new_date, new_date))
            files.append(filename)

        for filename in files:
            assert os.path.exists(filename), filename
        cleanup_directory(self.tmpdir, timestamp_before())
        for filename in files:
            assert not os.path.exists(filename), filename
            assert not os.path.exists(os.path.dirname(filename)), filename

    def test_remove_empty_dirs(self):
        os.makedirs(os.path.join(self.tmpdir, "foo", "bar", "baz"))
        cleanup_directory(self.tmpdir, timestamp_before(minutes=-1))
        assert not os.path.exists(os.path.join(self.tmpdir, "foo"))

    def test_remove_some(self):
        files = []
        # create a few files, every other file is one week old
        new_date = timestamp_before(weeks=1)
        for n in range(10):
            fname = "foo" + str(n)
            filename = self.mkfile(fname)
            if n % 2 == 0:
                os.utime(filename, (new_date, new_date))
            files.append(filename)

        # check all files are present
        for filename in files:
            assert os.path.exists(filename), filename

        # cleanup_directory for all files older then one minute
        cleanup_directory(self.tmpdir, timestamp_before(minutes=1))

        # check old files and dirs are removed
        for filename in files[::2]:
            assert not os.path.exists(filename), filename
            assert not os.path.exists(os.path.dirname(filename)), filename

        # check new files are still present
        for filename in files[1::2]:
            assert os.path.exists(filename), filename


def _write_atomic_data(i_filename):
    (i, filename) = i_filename
    data = str(i) + "\n" + "x" * 10000
    write_atomic(filename, data.encode("utf-8"))
    time.sleep(0.001)


class TestWriteAtomic(object):

    def setup_method(self):
        self.dirname = tempfile.mkdtemp()

    def teardown_method(self):
        if self.dirname:
            shutil.rmtree(self.dirname)

    def test_concurrent_write(self):
        filename = os.path.join(self.dirname, "tmpfile")

        num_writes = 800
        concurrent_writes = 8

        p = multiprocessing.Pool(concurrent_writes)
        p.map(_write_atomic_data, ((i, filename) for i in range(num_writes)))
        p.close()
        p.join()

        assert os.path.exists(filename)
        last_i = int(open(filename).readline())
        assert last_i > (num_writes / 2), (
            "file should contain content from "
            "later writes, got content from write %d" % (last_i + 1)
        )
        os.unlink(filename)
        assert os.listdir(self.dirname) == []

    def test_not_a_file(self):
        # check that expected errors are not hidden
        filename = os.path.join(self.dirname, "tmpfile")
        os.mkdir(filename)

        try:
            write_atomic(filename, b"12345")
        except (OSError, IOError):
            pass
        else:
            assert False, "expected exception"


def test_reraise_exception():

    def valueerror_raiser():
        raise ValueError()

    def reraiser():
        try:
            valueerror_raiser()
        except ValueError:
            reraise_exception(TypeError(), sys.exc_info())

    try:
        reraiser()
    except TypeError as ex:
        assert ex
    else:
        assert False, "expected exception"



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import mapproxy.service as service_package

from mapproxy.request.base import Request
from mapproxy.response import Response
from mapproxy.service.ogcapi.constants import (
    FORMAT_TYPES,
    F_JSON,
    F_HTML,
    MEDIA_TYPE_OPENAPI_3_0,
)
from mapproxy.service.ogcapi.server import OGCAPIServer
from mapproxy.util.jinja2_templates import render_j2_template
from mapproxy.version import __version__


def _get(d: dict, *path):
    for p in path:
        if p not in d:
            return None
        d = d[p]
    return d


def api(server: OGCAPIServer, req: Request):
    log = server.log
    log.debug("API page")

    for arg in req.args:
        if arg not in ("f", "ui"):
            raise OGCAPIServer.unknown_query_parameter(arg)
    if req.args.get("f", None) not in (None, F_JSON, F_HTML):
        raise OGCAPIServer.invalid_parameter("Invalid value for f query parameter")
    if req.args.get("ui", None) not in (None, "swagger", "redoc"):
        raise OGCAPIServer.invalid_parameter("Invalid value for ui query parameter")
    is_html_req = server.is_html_req(req)
    if req.args.get("ui", None) and not is_html_req:
        raise OGCAPIServer.invalid_parameter(
            "ui query parameter can only be used with HTML output"
        )

    cfg = server.get_pygeoapi_config(req)

    if is_html_req:
        template = "openapi/swagger.html"
        if req.args.get("ui", None) == "redoc":
            template = "openapi/redoc.html"

        path = server.create_href(req, "/ogcapi/api")
        data = {"openapi-document-path": path}
        content = render_j2_template(
            cfg, service_package.__package__, "ogcapi", template, data
        )
        return Response(
            content,
            content_type=FORMAT_TYPES[F_HTML],
            headers=server.response_headers,
            status=200,
        )

    locale_ = None

    oas = {"openapi": "3.0.2", "tags": []}
    info = {"version": __version__}

    title = _get(cfg, "metadata", "identification", "title")
    if title:
        info["title"] = title
    else:
        info["title"] = "OGCAPI implementation"

    description = _get(cfg, "metadata", "identification", "description")
    if description:
        info["description"] = description

    keywords = _get(cfg, "metadata", "identification", "keywords")
    if keywords:
        info["x-keywords"] = keywords

    terms_of_service = _get(cfg, "metadata", "identification", "terms_of_service")
    if terms_of_service:
        info["termsOfService"] = terms_of_service

    license = _get(cfg, "metadata", "license")
    if license:
        info["license"] = license

    if server.enable_maps and server.max_output_pixels:
        info["x-OGC-limits"] = {
            "maps": {
                "maxWidth": server.max_width,
                "maxHeight": server.max_height,
                "maxPixels": server.max_output_pixels,
            }
        }

    oas["info"] = info

    oas["servers"] = [
        {
            "url": cfg["server"]["url"],
        }
    ]
    server_description = _get(cfg, "metadata", "identification", "description")
    if server_description:
        oas["servers"][0]["description"] = server_description

    OPENAPI_YAML = {
        "oapim-1": "https://schemas.opengis.net/ogcapi/maps/part1/1.0/openapi/ogcapi-maps-1.bundled.json",  # noqa
        "oapit-1": "https://schemas.opengis.net/ogcapi/tiles/part1/1.0/openapi/ogcapi-tiles-1.bundled.json",  # noqa
    }

    paths = {}

    paths["/"] = {
        "get": {
            "summary": "Landing page",
            "description": "Landing page",
            "tags": ["server"],
            "operationId": "getLandingPage",
            "parameters": [
                {"$ref": "#/components/parameters/f"},
                # {'$ref': '#/components/parameters/lang'}
            ],
            "responses": {
                "200": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/LandingPage"
                },  # noqa
                "400": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/InvalidParameter"
                },  # noqa
                "500": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/ServerError"
                },  # noqa
            },
        }
    }

    paths["/api"] = {
        "get": {
            "summary": "This document",
            "description": "This document",
            "tags": ["server"],
            "operationId": "getOpenapi",
            "parameters": [
                {"$ref": "#/components/parameters/f"},
                # {'$ref': '#/components/parameters/lang'},
                {
                    "name": "ui",
                    "in": "query",
                    "description": "UI to render the OpenAPI document",
                    "required": False,
                    "schema": {
                        "type": "string",
                        "enum": ["swagger", "redoc"],
                        "default": "swagger",
                    },
                    "style": "form",
                    "explode": False,
                },
            ],
            "responses": {
                "200": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/API"
                },  #
                "400": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/InvalidParameter"
                },  # noqa
                "500": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/ServerError"
                },  # noqa
            },
        }
    }

    paths["/collections"] = {
        "get": {
            "summary": "Retrieve the description of the collections available from this service.",
            "description": "Collections",
            "tags": ["server"],
            "operationId": "getCollections",
            "parameters": [
                {"$ref": "#/components/parameters/f"},
                # {'$ref': '#/components/parameters/lang'}
            ],
            "responses": {
                "200": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/CollectionsList"
                },  # noqa
                "400": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/InvalidParameter"
                },  # noqa
                "500": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/ServerError"
                },  # noqa
            },
        }
    }

    paths["/collections/{collectionId}"] = {
        "get": {
            "summary": "Retrieve the description of a collection available from this service.",
            "description": "Collection",
            "tags": ["server"],
            "operationId": "getCollection",
            "parameters": [
                {"$ref": "#/components/parameters/collectionId"},
                {"$ref": "#/components/parameters/f"},
                # {'$ref': '#/components/parameters/lang'}
            ],
            "responses": {
                "200": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/Collection"
                },  # noqa
                "400": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/InvalidParameter"
                },  # noqa
                "404": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/NotFound"
                },  # noqa
                "500": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/ServerError"
                },  # noqa
            },
        }
    }

    if server.enable_maps:
        paths["/collections/{collectionId}/map"] = {
            "get": {
                "summary": "Retrieve the default map for the specified collection.",
                "description": "Map",
                "tags": ["server"],
                "operationId": "getCollectionMap",
                "parameters": [
                    {"$ref": "#/components/parameters/collectionId"},
                    {"$ref": "#/components/parameters/f_img"},
                    {"$ref": "#/components/parameters/width"},
                    {"$ref": "#/components/parameters/height"},
                    {"$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/crs"},
                    {"$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/bbox"},
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/bbox-crs"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/center"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/center-crs"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/subset"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/subset-crs"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/scale-denominator"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/bgcolor"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/transparent"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/mm-per-pixel"
                    },
                    # {'$ref': '#/components/parameters/lang'}
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/Map"
                    },  # noqa
                    "204": {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/EmptyMap"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "404": {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/NotFound"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }

    if server.enable_tiles:
        paths["/collections/{collectionId}/map/tiles"] = {
            "get": {
                "summary": "Retrieve a list of all map tilesets for specified collection",
                "description": "CollectionMapTileSets",
                "tags": ["server"],
                "operationId": "listCollectionMapTileSets",
                "parameters": [
                    {"$ref": "#/components/parameters/collectionId"},
                    {"$ref": "#/components/parameters/f"},
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/TileSetsList"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "404": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/NotFound"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }
        paths["/collections/{collectionId}/map/tiles/{tileMatrixSetId}"] = {
            "get": {
                "summary": "Retrieve a map tile set metadata for the specified collection and tiling scheme (tile matrix set)",  # noqa
                "description": "CollectionMapTileSet",
                "tags": ["server"],
                "operationId": "describeCollectionMapTileSet",
                "parameters": [
                    {"$ref": "#/components/parameters/collectionId"},
                    {"$ref": "#/components/parameters/tileMatrixSetId"},
                    {"$ref": "#/components/parameters/f"},
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/TileSet"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "404": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/NotFound"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }
        paths[
            "/collections/{collectionId}/map/tiles/{tileMatrixSetId}/{tileMatrix}/{tileRow}/{tileCol}"
        ] = {
            "get": {
                "summary": "Retrieve a map tile from the specified collection",
                "description": "CollectionMapTile",
                "tags": ["server"],
                "operationId": "getCollectionMapTile",
                "parameters": [
                    {"$ref": "#/components/parameters/collectionId"},
                    {"$ref": "#/components/parameters/tileMatrixSetId"},
                    {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/parameters/tileMatrix"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/parameters/tileRow"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/parameters/tileCol"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/bgcolor"
                    },
                    {"$ref": "#/components/parameters/f_img"},
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/MapTile"
                    },  # noqa
                    "204": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/EmptyTile"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "404": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/NotFound"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }

        if server.enable_maps:
            paths[
                "/collections/{collectionId}/map/tiles/{tileMatrixSetId}/{tileMatrix}/{tileRow}/{tileCol}"
            ]["get"]["parameters"] += [
                {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/scale-denominator"
                },
                {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/transparent"
                },
                {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/mm-per-pixel"
                },
                {"$ref": "#/components/parameters/width"},
                {"$ref": "#/components/parameters/height"},
            ]

    if server.enable_maps and server.default_dataset_layers is not None:
        paths["/map"] = {
            "get": {
                "summary": "Retrieve the default map for the whole dataset.",
                "description": "Map",
                "tags": ["server"],
                "operationId": "getDatasetMap",
                "parameters": [
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/collections"
                    },
                    {"$ref": "#/components/parameters/f_img"},
                    {"$ref": "#/components/parameters/width"},
                    {"$ref": "#/components/parameters/height"},
                    {"$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/crs"},
                    {"$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/bbox"},
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/bbox-crs"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/center"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/center-crs"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/subset"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/subset-crs"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/scale-denominator"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/bgcolor"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/transparent"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/mm-per-pixel"
                    },
                    # {'$ref': '#/components/parameters/lang'}
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/Map"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }

    if server.enable_tiles and server.default_dataset_layers is not None:
        paths["/map/tiles"] = {
            "get": {
                "summary": "Retrieve a list of all map tilesets for the wkole dataset",
                "description": "DatasetMapTileSets",
                "tags": ["server"],
                "operationId": "listDatasetMapTileSets",
                "parameters": [
                    {"$ref": "#/components/parameters/f"},
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/TileSetsList"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "404": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/NotFound"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }
        paths["/map/tiles/{tileMatrixSetId}"] = {
            "get": {
                "summary": "Retrieve a map tile set metadata for the whole datase and tiling scheme (tile matrix set)",  # noqa
                "description": "DatasetMapTileSet",
                "tags": ["server"],
                "operationId": "describeDatasetMapTileSet",
                "parameters": [
                    {"$ref": "#/components/parameters/tileMatrixSetId"},
                    {"$ref": "#/components/parameters/f"},
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/TileSet"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "404": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/NotFound"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }
        paths["/map/tiles/{tileMatrixSetId}/{tileMatrix}/{tileRow}/{tileCol}"] = {
            "get": {
                "summary": "Retrieve a map tile from the whole dataset",
                "description": "CollectionMapTile",
                "tags": ["server"],
                "operationId": "getDatasetMapTile",
                "parameters": [
                    {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/parameters/collections"
                    },
                    {"$ref": "#/components/parameters/tileMatrixSetId"},
                    {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/parameters/tileMatrix"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/parameters/tileRow"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/parameters/tileCol"
                    },
                    {
                        "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/bgcolor"
                    },
                    {"$ref": "#/components/parameters/f_img"},
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/MapTile"
                    },  # noqa
                    "204": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/EmptyTile"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "404": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/NotFound"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }

        if server.enable_maps:
            paths["/map/tiles/{tileMatrixSetId}/{tileMatrix}/{tileRow}/{tileCol}"][
                "get"
            ]["parameters"] += [
                {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/scale-denominator"
                },
                {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/transparent"
                },
                {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/parameters/mm-per-pixel"
                },
                {"$ref": "#/components/parameters/width"},
                {"$ref": "#/components/parameters/height"},
            ]

    if server.enable_tiles and server.grid_configs:
        paths["/tileMatrixSets"] = {
            "get": {
                "tags": ["server"],
                "description": "TileMatrixSets",
                "summary": "Retrieve the list of available tiling schemes (tile matrix sets)",
                "operationId": "listTileMatrixSets",
                "parameters": [
                    {"$ref": "#/components/parameters/f"},
                    # {'$ref': '#/components/parameters/lang'}
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/TileMatrixSetsList"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }

        paths["/tileMatrixSets/{tileMatrixSetId}"] = {
            "get": {
                "tags": ["server"],
                "description": "TileMatrixSet",
                "summary": "Retrieve the definition of the specified tiling scheme (tile matrix set)",
                "operationId": "describeTileMatrixSet",
                "parameters": [
                    {"$ref": "#/components/parameters/f"},
                    {"$ref": "#/components/parameters/tileMatrixSetId"},
                    # {'$ref': '#/components/parameters/lang'}
                ],
                "responses": {
                    "200": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/TileMatrixSet"
                    },  # noqa
                    "400": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/InvalidParameter"
                    },  # noqa
                    "500": {
                        "$ref": f"{OPENAPI_YAML['oapit-1']}#/components/responses/ServerError"
                    },  # noqa
                },
            }
        }
    paths["/conformance"] = {
        "get": {
            "summary": "API conformance definition",
            "description": "API conformance definition",
            "tags": ["server"],
            "operationId": "getConformanceDeclaration",
            "parameters": [
                {"$ref": "#/components/parameters/f"},
                # {'$ref': '#/components/parameters/lang'}
            ],
            "responses": {
                "200": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/Conformance"
                },  # noqa
                "400": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/InvalidParameter"
                },  # noqa
                "500": {
                    "$ref": f"{OPENAPI_YAML['oapim-1']}#/components/responses/ServerError"
                },  # noqa
            },
        }
    }

    oas["paths"] = paths

    tags_server = {
        "name": "server",
        "description": description if description else "OGCAPI implementation",  # noqa
    }
    identification_url = _get(cfg, "metadata", "identification", "url")
    if identification_url:
        tags_server["externalDocs"] = {
            "description": "information",
            "url": identification_url,
        }
    oas["tags"].append(tags_server)

    oas["components"] = {
        "responses": {
            "200": {"description": "Successful operation"},
        },
        "parameters": _get_openapi_parameters(server, cfg=cfg, locale_=locale_),
    }

    return Response(
        json.dumps(oas, indent=2),
        content_type=MEDIA_TYPE_OPENAPI_3_0,
        headers=server.response_headers,
        status=200,
    )


def _get_openapi_parameters(server: OGCAPIServer, cfg: dict, locale_: str):
    parameters = {
        "f": {
            "name": "f",
            "in": "query",
            "description": "The optional f parameter indicates the output format which the server shall provide as part of the response document.  The default format is JSON.",  # noqa
            "required": False,
            "schema": {
                "type": "string",
                "enum": ["json", "html"],
                "default": "json",
            },
            "style": "form",
            "explode": False,
        },
        "collectionId": {
            "name": "collectionId",
            "in": "path",
            "description": "Local identifier of a collection",
            "required": True,
            "allowEmptyValue": False,
            "schema": {
                "type": "string",
                "enum": [id for id in server.layers],
            },
        },
    }

    if server.enable_tiles and server.grid_configs:
        parameters["tileMatrixSetId"] = {
            "name": "tileMatrixSetId",
            "in": "path",
            "description": "Identifier for a supported TileMatrixSet",
            "required": True,
            "allowEmptyValue": False,
            "schema": {
                "type": "string",
                "enum": [id for id in server.grid_configs],
            },
        }

    if server.enable_maps:
        parameters["f_img"] = {
            "name": "f",
            "in": "query",
            "description": "The optional f parameter indicates the output format which the server shall provide as part of the response document.  The default format is PNG.",  # noqa
            "required": False,
            "schema": {
                "type": "string",
                "enum": ["png", "jpeg"],
                "default": "png",
            },
            "style": "form",
            "explode": False,
        }

        parameters["width"] = {
            "name": "width",
            "in": "query",
            "description": "Width of the map in pixels. If omitted and `height` is specified, defaults to the width maintaining a 1:1 aspect ratio. If both `width` and `height` are omitted, the server will select default dimensions. When used together with the `center` and/or `scale-denominator` parameter, `width` takes on a subsetting role rather than scaling (resampling), defining the horizontal portion of the map to subset based on the scale (native scale, or specified by `scale-denominator`) and display resolution (0.28 mm/pixel, or specified by `mm-per-pixel`).",  # noqa
            "required": False,
            "style": "form",
            "explode": False,
            "schema": {"type": "integer"},
        }

        parameters["height"] = {
            "name": "height",
            "in": "query",
            "description": "Height of the map in pixels. If omitted and `width` is specified, defaults to the height maintaining a 1:1 aspect ratio. If both `width` and `height` are omitted, the server will select default dimensions. When used together with the `center` and/or `scale-denominator` parameter, `height` takes on a subsetting role rather than scaling (resampling), defining the vertical portion of the map to subset based on the scale (native scale, or specified by `scale-denominator`) and display resolution (0.28 mm/pixel, or specified by `mm-per-pixel`).",  # noqa
            "required": False,
            "style": "form",
            "explode": False,
            "schema": {"type": "integer"},
        }

        if server.max_width:
            parameters["width"]["schema"]["maximum"] = server.max_width

        if server.max_height:
            parameters["height"]["schema"]["maximum"] = server.max_height

    return parameters



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from io import BytesIO

import pytest

from mapproxy.request.wms import WMS111FeatureInfoRequest
from mapproxy.test.image import is_png, create_tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "arcgis.yaml"


transp = create_tmp_image((512, 512), mode="RGBA", color=(0, 0, 0, 0))


class TestArcgisSource(SysTest):

    def setup_method(self):
        self.common_fi_req = WMS111FeatureInfoRequest(
            url="/service?",
            param=dict(
                x="10",
                y="20",
                width="200",
                height="200",
                layers="app2_with_layers_fi_layer",
                format="image/png",
                query_layers="app2_with_layers_fi_layer",
                styles="",
                bbox="1000,400,2000,1400",
                srs="EPSG:3857",
                info_format="application/json",
            ),
        )

    def test_get_tile(self, app):
        expected_req = [
            (
                {
                    "path": "/arcgis/rest/services/ExampleLayer/ImageServer/exportImage?f=image&format=png&imageSR=900913&bboxSR=900913&bbox=-20037508.342789244,-20037508.342789244,20037508.342789244,20037508.342789244&size=512,512"  # noqa
                },
                {"body": transp, "headers": {"content-type": "image/png"}},
            )
        ]

        with mock_httpd(
            ("localhost", 42423), expected_req, bbox_aware_query_comparator=True
        ):
            resp = app.get("/tms/1.0.0/app2_layer/0/0/1.png")
            assert resp.content_type == "image/png"
            assert resp.content_length == len(resp.body)
            data = BytesIO(resp.body)
            assert is_png(data)

    def test_get_tile_with_layer(self, app):
        expected_req = [
            (
                {
                    "path": "/arcgis/rest/services/ExampleLayer/MapServer/export?f=image&format=png&layers=show:0,1&imageSR=900913&bboxSR=900913&bbox=-20037508.342789244,-20037508.342789244,20037508.342789244,20037508.342789244&size=512,512"  # noqa
                },
                {"body": transp, "headers": {"content-type": "image/png"}},
            )
        ]

        with mock_httpd(
            ("localhost", 42423), expected_req, bbox_aware_query_comparator=True
        ):
            resp = app.get("/tms/1.0.0/app2_with_layers_layer/0/0/1.png")
            assert resp.content_type == "image/png"
            assert resp.content_length == len(resp.body)
            data = BytesIO(resp.body)
            assert is_png(data)

    def test_get_tile_from_missing_arcgis_layer(self, app):
        expected_req = [
            (
                {
                    "path": "/arcgis/rest/services/NonExistentLayer/ImageServer/exportImage?f=image&format=png&imageSR=900913&bboxSR=900913&bbox=-20037508.342789244,-20037508.342789244,20037508.342789244,20037508.342789244&size=512,512"  # noqa
                },
                {"body": b"", "status": 400},
            )
        ]

        with mock_httpd(
            ("localhost", 42423), expected_req, bbox_aware_query_comparator=True
        ):
            resp = app.get("/tms/1.0.0/app2_wrong_url_layer/0/0/1.png", status=500)
            assert resp.status_code == 500

    def test_identify(self, app):
        expected_req = [
            (
                {
                    "path": "/arcgis/rest/services/ExampleLayer/MapServer/identify?f=json&"
                    "geometry=1050.000000,1300.000000&returnGeometry=true&imageDisplay=200,200,96"
                    "&mapExtent=1000.0,400.0,2000.0,1400.0&layers=show:1,2,3"
                    "&tolerance=10&geometryType=esriGeometryPoint&sr=3857"
                },
                {
                    "body": b'{"results": []}',
                    "headers": {"content-type": "application/json"},
                },
            )
        ]

        with mock_httpd(
            ("localhost", 42423), expected_req, bbox_aware_query_comparator=True
        ):
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "application/json"
            assert resp.content_length == len(resp.body)
            assert resp.body == b'{"results": []}'

    def test_transformed_identify(self, app):
        expected_req = [
            (
                {
                    "path": "/arcgis/rest/services/ExampleLayer/MapServer/identify?f=json&"
                    "geometry=573295.377585,6927820.884193&returnGeometry=true&imageDisplay=200,321,96"
                    "&mapExtent=556597.453966,6446275.84102,890555.926346,6982997.92039&layers=show:1,2,3"
                    "&tolerance=10&geometryType=esriGeometryPoint&sr=3857"
                },
                {
                    "body": b'{"results": []}',
                    "headers": {"content-type": "application/json"},
                },
            )
        ]

        with mock_httpd(("localhost", 42423), expected_req):
            self.common_fi_req.params.bbox = "5,50,8,53"
            self.common_fi_req.params.srs = "EPSG:4326"
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "application/json"
            assert resp.content_length == len(resp.body)
            assert resp.body == b'{"results": []}'



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from contextlib import contextmanager
from io import BytesIO

import pytest

from PIL import Image, ImageDraw, ImageColor
from mapproxy.request.wms import WMS111MapRequest
from mapproxy.request.wmts import WMTS100TileRequest
from mapproxy.test.image import check_format, is_transparent
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "mixed_mode.yaml"


class TestWMS(SysTest):

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="0,0,180,80",
                width="200",
                height="200",
                layers="mixed_mode",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                transparent="true",
            ),
        )
        self.expected_base_path = "/service?SERVICE=WMS&REQUEST=GetMap&HEIGHT=256"\
                                  "&SRS=EPSG%3A900913&styles=&VERSION=1.1.1&WIDTH=512"\
                                  "&BBOX=-20037508.3428,0.0,20037508.3428,20037508.3428"

    def test_mixed_mode(self, app, cache_dir):
        req_format = "png"
        transparent = "True"
        with create_mixed_mode_img((512, 256)) as img:
            expected_req = (
                {
                    "path": self.expected_base_path
                    + "&layers=mixedsource"
                    + "&format=image%2F"
                    + req_format
                    + "&transparent="
                    + transparent
                },
                {
                    "body": img.read(),
                    "headers": {"content-type": "image/" + req_format},
                },
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["format"] = "image/" + req_format
                resp = app.get(self.common_map_req)

                assert resp.content_type == "image/" + req_format
                check_format(BytesIO(resp.body), req_format)
                # GetMap Request is fully within the opaque tile
                assert not is_transparent(resp.body)

        # check cache formats
        for f, format in [
            ["mixed_cache_EPSG900913/01/000/000/000/000/000/001.mixed", "png"],
            ["mixed_cache_EPSG900913/01/000/000/001/000/000/001.mixed", "jpeg"],
        ]:
            assert cache_dir.join(f).check()
            check_format(cache_dir.join(f).read_binary(), format)


class TestTMS(SysTest):

    def setup_method(self):
        self.expected_base_path = "/service?SERVICE=WMS&REQUEST=GetMap&HEIGHT=256"\
                                  "&SRS=EPSG%3A900913&styles=&VERSION=1.1.1&WIDTH=512"\
                                  "&BBOX=-20037508.3428,-20037508.3428,20037508.3428,0.0"

    def test_mixed_mode(self, app, cache_dir):
        with create_mixed_mode_img((512, 256)) as img:
            expected_req = (
                {
                    "path": self.expected_base_path
                    + "&layers=mixedsource"
                    + "&format=image%2Fpng"
                    + "&transparent=True"
                },
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get("/tms/1.0.0/mixed_mode/0/0/0.png")
                assert resp.content_type == "image/png"
                assert is_transparent(resp.body)

                resp = app.get("/tms/1.0.0/mixed_mode/0/1/0.png")

                assert resp.content_type == "image/jpeg"
        assert cache_dir.join(
            "mixed_cache_EPSG900913/01/000/000/000/000/000/000.mixed"
        ).check()
        assert cache_dir.join(
            "mixed_cache_EPSG900913/01/000/000/001/000/000/000.mixed"
        ).check()


class TestWMTS(SysTest):

    def setup_method(self):
        self.common_tile_req = WMTS100TileRequest(
            url="/service?",
            param=dict(
                service="WMTS",
                version="1.0.0",
                tilerow="0",
                tilecol="0",
                tilematrix="01",
                tilematrixset="GLOBAL_MERCATOR",
                layer="mixed_mode",
                format="image/png",
                style="",
                request="GetTile",
                transparent="True",
            ),
        )
        self.expected_base_path = "/service?SERVICE=WMS&REQUEST=GetMap&HEIGHT=256"\
                                  "&SRS=EPSG%3A900913&styles=&VERSION=1.1.1&WIDTH=512"\
                                  "&BBOX=-20037508.3428,0.0,20037508.3428,20037508.3428"

    def test_mixed_mode(self, app, cache_dir):
        with create_mixed_mode_img((512, 256)) as img:
            expected_req = (
                {
                    "path": self.expected_base_path
                    + "&layers=mixedsource"
                    + "&format=image%2Fpng"
                    + "&transparent=True"
                },
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get(self.common_tile_req)
                assert resp.content_type == "image/png"
                assert is_transparent(resp.body)

                self.common_tile_req.params["tilecol"] = "1"
                resp = app.get(self.common_tile_req)
                assert resp.content_type == "image/jpeg"
        assert cache_dir.join(
            "mixed_cache_EPSG900913/01/000/000/000/000/000/001.mixed"
        ).check()
        assert cache_dir.join(
            "mixed_cache_EPSG900913/01/000/000/001/000/000/001.mixed"
        ).check()


@contextmanager
def create_mixed_mode_img(size, format="png"):
    img = Image.new("RGBA", size)

    # draw a black rectangle into the image, rect_width = 3/4 img_width
    # thus 1/4 of the image is transparent and with square tiles, one of two
    # tiles (img size = 512x256) will be fully opaque and the other
    # has transparency
    draw = ImageDraw.Draw(img)
    w, h = size
    red_color = ImageColor.getrgb("red")
    draw.rectangle((w / 4, 0, w, h), fill=red_color)

    data = BytesIO()
    img.save(data, format)
    data.seek(0)
    yield data



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
import math

import pytest

from mapproxy.request.wms import (
    WMS111MapRequest,
    WMS111CapabilitiesRequest,
    WMS130CapabilitiesRequest,
)

from mapproxy.test.system import SysTest
from mapproxy.test.image import is_png, is_transparent, tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.system.test_wms import is_111_capa, is_130_capa, ns130


@pytest.fixture(scope="module")
def config_file():
    return "scalehints.yaml"


def diagonal_res_to_pixel_res(res):
    """
    >>> '%.2f' % round(diagonal_res_to_pixel_res(14.14214), 4)
    '10.00'
    """
    return math.sqrt((float(res) ** 2) / 2)


class TestWMS(SysTest):

    def setup_method(self):
        self.common_req = WMS111MapRequest(
            url="/service?", param=dict(service="WMS", version="1.1.1")
        )
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="res",
                srs="EPSG:4326",
                format="image/png",
                transparent="true",
                styles="",
                request="GetMap",
            ),
        )

    def test_capabilities_111(self, app):
        req = WMS111CapabilitiesRequest(url="/service?").copy_with_request_params(
            self.common_req
        )
        resp = app.get(req)
        xml = resp.lxml
        assert is_111_capa(xml)
        hints = xml.xpath("//Layer/Layer/ScaleHint")
        assert diagonal_res_to_pixel_res(hints[0].attrib["min"]) == pytest.approx(10)
        assert diagonal_res_to_pixel_res(hints[0].attrib["max"]) == pytest.approx(10000)

        assert diagonal_res_to_pixel_res(hints[1].attrib["min"]) == pytest.approx(2.8)
        assert diagonal_res_to_pixel_res(hints[1].attrib["max"]) == pytest.approx(280)

        assert diagonal_res_to_pixel_res(hints[2].attrib["min"]) == pytest.approx(0.28)
        assert diagonal_res_to_pixel_res(hints[2].attrib["max"]) == pytest.approx(2.8)

    def test_capabilities_130(self, app):
        req = WMS130CapabilitiesRequest(url="/service?").copy_with_request_params(
            self.common_req
        )
        resp = app.get(req)
        xml = resp.lxml
        assert is_130_capa(xml)
        min_scales = xml.xpath(
            "//wms:Layer/wms:Layer/wms:MinScaleDenominator/text()", namespaces=ns130
        )
        max_scales = xml.xpath(
            "//wms:Layer/wms:Layer/wms:MaxScaleDenominator/text()", namespaces=ns130
        )

        assert float(min_scales[0]) == pytest.approx(35714.28)
        assert float(max_scales[0]) == pytest.approx(35714285.7)

        assert float(min_scales[1]) == pytest.approx(10000)
        assert float(max_scales[1]) == pytest.approx(1000000)

        assert float(min_scales[2]) == pytest.approx(1000)
        assert float(max_scales[2]) == pytest.approx(10000)

    def test_get_map_above_res(self, app):
        # no layer rendered
        resp = app.get(self.common_map_req)
        assert is_png(resp.body)
        assert is_transparent(resp.body)

    def test_get_map_mixed(self, app, cache_dir):
        # only res layer matches resolution range
        self.common_map_req.params["layers"] = "res,scale"
        self.common_map_req.params["bbox"] = "0,0,100000,100000"
        self.common_map_req.params["srs"] = "EPSG:900913"
        self.common_map_req.params.size = 100, 100
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=reslayer&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=0.0,0.0,156543.033928,156543.033928"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                resp = app.get(self.common_map_req)
                assert is_png(resp.body)
                assert not is_transparent(resp.body)

        assert cache_dir.join(
            "res_cache_EPSG900913/08/000/000/128/000/000/128.jpeg"
        ).check()



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO

import pytest

from PIL import Image
from mapproxy.image import ImageSource
from mapproxy.request.wms import WMS111MapRequest
from mapproxy.request.wmts import WMTS100TileRequest
from mapproxy.test.image import is_png, is_jpeg
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "layer.yaml"


def to_greyscale(image, service, layers, **kw):
    img = image.as_image()
    if hasattr(image.image_opts, "transparent") and image.image_opts.transparent:
        img = img.convert("LA").convert("RGBA")
    else:
        img = img.convert("L").convert("RGB")
    return ImageSource(img, image.image_opts)


@pytest.mark.usefixtures("fixture_cache_data")
class TestDecorateImg(SysTest):

    def setup_method(self):
        self.common_tile_req = WMTS100TileRequest(
            url="/service?",
            param=dict(
                service="WMTS",
                version="1.0.0",
                tilerow="0",
                tilecol="0",
                tilematrix="01",
                tilematrixset="GLOBAL_MERCATOR",
                layer="wms_cache",
                format="image/jpeg",
                style="",
                request="GetTile",
            ),
        )

    def test_wms(self, app):
        req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )
        resp = app.get(req, extra_environ={"mapproxy.decorate_img": to_greyscale})
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGB"

    def test_wms_transparent(self, app):
        req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache_transparent",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                transparent="True",
            ),
        )
        resp = app.get(req, extra_environ={"mapproxy.decorate_img": to_greyscale})
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGBA"

    def test_wms_bgcolor(self, app):
        req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache_transparent",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                bgcolor="0xff00a0",
            ),
        )
        resp = app.get(req, extra_environ={"mapproxy.decorate_img": to_greyscale})
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGB"
        assert sorted(img.getcolors())[-1][1] == (94, 94, 94)

    def test_wms_args(self, app):
        req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache,wms_cache_transparent",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
                transparent="True",
            ),
        )

        def callback(img_src, service, layers, environ, query_extent):
            assert isinstance(img_src, ImageSource)
            assert "wms.map" == service
            assert len(layers) == 2
            assert "wms_cache_transparent" in layers
            assert "wms_cache" in layers
            assert isinstance(environ, dict)
            assert len(query_extent) == 2
            assert len(query_extent[1]) == 4
            assert query_extent[0] == "EPSG:4326"
            return img_src

        app.get(req, extra_environ={"mapproxy.decorate_img": callback})

    def test_tms(self, app):
        resp = app.get(
            "/tms/1.0.0/wms_cache/0/0/1.jpeg",
            extra_environ={"mapproxy.decorate_img": to_greyscale},
        )
        assert resp.content_type == "image/jpeg"
        assert resp.content_length == len(resp.body)
        data = BytesIO(resp.body)
        assert is_jpeg(data)

    def test_tms_args(self, app):

        def callback(img_src, service, layers, environ, query_extent):
            assert isinstance(img_src, ImageSource)
            assert "tms" == service
            assert "wms_cache" == layers[0]
            assert isinstance(environ, dict)
            assert len(query_extent) == 2
            assert len(query_extent[1]) == 4
            assert query_extent[0] == "EPSG:900913"
            return img_src

        app.get(
            "/tms/1.0.0/wms_cache/0/0/1.jpeg",
            extra_environ={"mapproxy.decorate_img": callback},
        )

    def test_wmts(self, app):
        resp = app.get(
            str(self.common_tile_req),
            extra_environ={"mapproxy.decorate_img": to_greyscale},
        )
        assert resp.content_type == "image/jpeg"
        assert resp.content_length == len(resp.body)
        data = BytesIO(resp.body)
        assert is_jpeg(data)

    def test_wmts_args(self, app):

        def callback(img_src, service, layers, environ, query_extent):
            assert isinstance(img_src, ImageSource)
            assert "wmts" == service
            assert "wms_cache" == layers[0]
            assert isinstance(environ, dict)
            assert len(query_extent) == 2
            assert len(query_extent[1]) == 4
            assert query_extent[0] == "EPSG:900913"
            return img_src

        app.get(
            str(self.common_tile_req), extra_environ={"mapproxy.decorate_img": callback}
        )



from __future__ import print_function
import math

from .util import resolve_ns

from xml.etree import ElementTree as etree
from mapproxy.request.wms import switch_bbox_epsg_axis_order


class WMSCapabilities(object):
    _default_namespace = None
    _namespaces = {
        'xlink': 'http://www.w3.org/1999/xlink',
    }

    version = None

    def __init__(self, tree):
        self.tree = tree
        self._layer_tree = None

    def resolve_ns(self, xpath):
        return resolve_ns(xpath, self._namespaces, self._default_namespace)

    def findtext(self, tree, xpath):
        return tree.findtext(self.resolve_ns(xpath))

    def find(self, tree, xpath):
        return tree.find(self.resolve_ns(xpath))

    def findall(self, tree, xpath):
        return tree.findall(self.resolve_ns(xpath))

    def attrib(self, elem, name):
        return elem.attrib[self.resolve_ns(name)]

    def metadata(self):
        md = dict(
            name=self.findtext(self.tree, 'Service/Name'),
            title=self.findtext(self.tree, 'Service/Title'),
            abstract=self.findtext(self.tree, 'Service/Abstract'),
            fees=self.findtext(self.tree, 'Service/Fees'),
            access_constraints=self.findtext(self.tree, 'Service/AccessConstraints'),
        )
        elem = self.find(self.tree, 'Service/OnlineResource')
        if elem is not None:
            md['online_resource'] = self.attrib(elem, 'xlink:href')

        md['contact'] = self.parse_contact()
        return md

    def parse_contact(self):
        elem = self.find(self.tree, 'Service/ContactInformation')
        if elem is None or len(elem) == 0:
            elem = etree.Element(None)
        md = dict(
            person=self.findtext(elem, 'ContactPersonPrimary/ContactPerson'),
            organization=self.findtext(elem, 'ContactPersonPrimary/ContactOrganization'),
            position=self.findtext(elem, 'ContactPosition'),

            address=self.findtext(elem, 'ContactAddress/Address'),
            city=self.findtext(elem, 'ContactAddress/City'),
            postcode=self.findtext(elem, 'ContactAddress/PostCode'),
            country=self.findtext(elem, 'ContactAddress/Country'),
            phone=self.findtext(elem, 'ContactVoiceTelephone'),
            fax=self.findtext(elem, 'ContactFacsimileTelephone'),
            email=self.findtext(elem, 'ContactElectronicMailAddress'),
        )

        return md

    def layers(self):
        if not self._layer_tree:
            root_layer = self.find(self.tree, 'Capability/Layer')
            self._layer_tree = self.parse_layer(root_layer, None)

        return self._layer_tree

    def layers_list(self):
        layers = []

        def append_layer(layer):
            if layer.get('name'):
                layers.append(layer)
            for child_layer in layer.get('layers', []):
                append_layer(child_layer)

        append_layer(self.layers())
        return layers

    def requests(self):
        requests_elem = self.find(self.tree, 'Capability/Request')
        resources = {}
        resource = self.find(requests_elem, 'GetMap/DCPType/HTTP/Get/OnlineResource')
        if resource is not None:
            resources['GetMap'] = self.attrib(resource, 'xlink:href')
        return resources

    def parse_layer(self, layer_elem, parent_layer):
        child_layers = []
        layer = self.parse_layer_data(layer_elem, parent_layer or {})
        child_layer_elems = self.findall(layer_elem, 'Layer')

        for child_elem in child_layer_elems:
            child_layers.append(self.parse_layer(child_elem, layer))

        layer['layers'] = child_layers
        return layer

    def parse_layer_data(self, elem, parent_layer):
        layer = dict(
            queryable=elem.attrib.get('queryable') == '1',
            opaque=elem.attrib.get('opaque') == '1',
            title=self.findtext(elem, 'Title'),
            abstract=self.findtext(elem, 'Abstract'),
            name=self.findtext(elem, 'Name'),
        )

        layer['srs'] = self.layer_srs(elem, parent_layer)
        layer['res_hint'] = self.layer_res_hint(elem, parent_layer)
        layer['llbbox'] = self.layer_llbbox(elem, parent_layer)
        layer['bbox_srs'] = self.layer_bbox_srs(elem, parent_layer)
        layer['url'] = self.requests()['GetMap']
        layer['legend'] = self.layer_legend(elem)

        return layer

    def layer_legend(self, elem):
        style_elems = self.findall(elem, 'Style')
        legend_elem = None
        # we don't support styles, but will use the
        # LegendURL for the default style
        for elem in style_elems:
            if self.findtext(elem, 'Name') in ('default', ''):
                legend_elem = self.find(elem, 'LegendURL')
                break

        if legend_elem is None:
            return

        legend = {}
        legend_url = self.find(legend_elem, 'OnlineResource')
        legend['url'] = self.attrib(legend_url, 'xlink:href')
        return legend

    def layer_res_hint(self, elem, parent_layer):
        elem = self.find(elem, 'ScaleHint')
        if elem is None:
            return parent_layer.get('res_hint')
        # ScaleHints are the diagonal pixel resolutions
        # NOTE: max is not the maximum resolution, but the max
        # value, so it's actualy the min_res
        min_res = elem.attrib.get('max')
        max_res = elem.attrib.get('min')

        if min_res:
            min_res = math.sqrt(float(min_res) ** 2 / 2.0)
        if max_res:
            max_res = math.sqrt(float(max_res) ** 2 / 2.0)

        return min_res, max_res


class WMS111Capabilities(WMSCapabilities):
    version = '1.1.1'

    def layer_llbbox(self, elem, parent_layer):
        llbbox_elem = self.find(elem, 'LatLonBoundingBox')
        llbbox = None
        if llbbox_elem is not None:
            llbbox = (
                llbbox_elem.attrib['minx'],
                llbbox_elem.attrib['miny'],
                llbbox_elem.attrib['maxx'],
                llbbox_elem.attrib['maxy']
            )
            llbbox = [float(x) for x in llbbox]
        elif parent_layer and 'llbbox' in parent_layer:
            llbbox = parent_layer['llbbox']
        return llbbox

    def layer_srs(self, elem, parent_layer=None):
        srs_elements = self.findall(elem, 'SRS')
        srs_codes = set()

        for srs in srs_elements:
            srs = srs.text.strip().upper()
            if ' ' in srs:
                # handle multiple codes in one SRS tag (WMS 1.1.1 7.1.4.5.5)
                srs_codes.update(srs.split())
            else:
                srs_codes.add(srs)

        # unique srs-codes in either srs or parent_layer['srs']
        inherited_srs = parent_layer.get('srs', set()) if parent_layer else set()
        return srs_codes | inherited_srs

    def layer_bbox_srs(self, elem, parent_layer=None):
        bbox_srs = {}

        bbox_srs_elems = self.findall(elem, 'BoundingBox')
        if len(bbox_srs_elems) > 0:
            for bbox_srs_elem in bbox_srs_elems:
                srs = bbox_srs_elem.attrib['SRS']
                bbox = (
                    bbox_srs_elem.attrib['minx'],
                    bbox_srs_elem.attrib['miny'],
                    bbox_srs_elem.attrib['maxx'],
                    bbox_srs_elem.attrib['maxy']
                )
                bbox = [float(x) for x in bbox]
                bbox_srs[srs] = bbox
        elif parent_layer:
            bbox_srs = parent_layer['bbox_srs']

        return bbox_srs


class WMS130Capabilities(WMSCapabilities):
    version = '1.3.0'
    _default_namespace = 'http://www.opengis.net/wms'
    _ns = {
        'sld': "http://www.opengis.net/sld",
        'xlink': "http://www.w3.org/1999/xlink",
    }

    def layer_llbbox(self, elem, parent_layer):
        llbbox_elem = self.find(elem, 'EX_GeographicBoundingBox')
        llbbox = None
        if llbbox_elem is not None:
            llbbox = (
                self.find(llbbox_elem, 'westBoundLongitude').text,
                self.find(llbbox_elem, 'southBoundLatitude').text,
                self.find(llbbox_elem, 'eastBoundLongitude').text,
                self.find(llbbox_elem, 'northBoundLatitude').text
            )
            llbbox = [float(x) for x in llbbox]
        elif parent_layer and 'llbbox' in parent_layer:
            llbbox = parent_layer['llbbox']

        return llbbox

    def layer_srs(self, elem, parent_layer=None):
        srs_elements = self.findall(elem, 'CRS')
        srs_codes = set([srs.text.strip().upper() for srs in srs_elements])
        # unique srs-codes in either srs or parent_layer['srs']
        inherited_srs = parent_layer.get('srs', set()) if parent_layer else set()
        return srs_codes | inherited_srs

    def layer_bbox_srs(self, elem, parent_layer=None):
        bbox_srs = {}

        bbox_srs_elems = self.findall(elem, 'BoundingBox')
        if len(bbox_srs_elems) > 0:
            for bbox_srs_elem in bbox_srs_elems:
                srs = bbox_srs_elem.attrib['CRS']
                bbox = (
                    bbox_srs_elem.attrib['minx'],
                    bbox_srs_elem.attrib['miny'],
                    bbox_srs_elem.attrib['maxx'],
                    bbox_srs_elem.attrib['maxy']
                )
                bbox = [float(x) for x in bbox]
                bbox = switch_bbox_epsg_axis_order(bbox, srs)
                bbox_srs[srs] = bbox
        elif parent_layer:
            bbox_srs = parent_layer['bbox_srs']

        return bbox_srs


def yaml_sources(cap):
    sources = {}
    for layer in cap.layers():
        layer_name = layer['name'] + '_wms'
        req = dict(url='http://example', layers=layer['name'])
        if not layer['opaque']:
            req['transparent'] = True

        sources[layer_name] = dict(
            type='wms',
            req=req
        )

    import yaml
    print(yaml.dump(dict(sources=sources), default_flow_style=False))


def parse_capabilities(fileobj):
    if isinstance(fileobj, str):
        fileobj = open(fileobj, 'rb')
    tree = etree.parse(fileobj)
    root_tag = tree.getroot().tag
    if root_tag == 'WMT_MS_Capabilities':
        return WMS111Capabilities(tree)
    elif root_tag == '{http://www.opengis.net/wms}WMS_Capabilities':
        return WMS130Capabilities(tree)
    elif root_tag.startswith('{http://www.opengis.net/wmts'):
        raise ValueError('Parsing of WMTS capabilities is not supported')
    else:
        raise ValueError(f'Unknown start tag in capabilities: {root_tag}')


if __name__ == '__main__':
    import sys
    cap = parse_capabilities(sys.argv[1])
    yaml_sources(cap)



# -:- encoding: utf-8 -:-
# This file is part of the MapProxy project.
# Copyright (C) 2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def caches(cap, sources, srs_grids):
    caches = {}
    for name, source in sources.items():
        conf = for_source(name, source, srs_grids)
        if not conf:
            continue
        caches[name[:-len('_wms')] + '_cache'] = conf

    return caches


def for_source(name, source, srs_grids):
    cache = {
        'sources': [name]
    }

    grids = []
    for srs in source['supported_srs']:
        if srs in srs_grids:
            grids.append(srs_grids[srs])

    if not grids:
        return None

    cache['grids'] = grids

    return cache



# This file is part of the MapProxy project.
# Copyright (C) 2010-2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from mapproxy.test.image import is_jpeg
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "tileservice_origin.yaml"


@pytest.mark.usefixtures("fixture_cache_data")
class TestTileServicesOrigin(SysTest):

    ###
    # tile 0/0/1 is cached, check if we can access it with different URLs

    def test_get_cached_tile_tms(self, app):
        resp = app.get("/tms/1.0.0/wms_cache/0/0/1.jpeg")
        assert resp.content_type == "image/jpeg"
        assert is_jpeg(resp.body)

    def test_get_cached_tile_service_origin(self, app):
        resp = app.get("/tiles/wms_cache/1/0/0.jpeg")
        assert resp.content_type == "image/jpeg"
        assert is_jpeg(resp.body)

    def test_get_cached_tile_request_origin(self, app):
        resp = app.get("/tiles/wms_cache/1/0/1.jpeg?origin=sw")
        assert resp.content_type == "image/jpeg"
        assert is_jpeg(resp.body)
